"""
Module for controlling the LED matrix or reading environment data on the SenseHat of a Raspberry Pi.

.. versionadded:: 2017.7.0

:maintainer:    Benedikt Werner <1benediktwerner@gmail.com>, Joachim Werner <joe@suse.com>
:maturity:      new
:depends:       sense_hat Python module

The rotation of the Pi can be specified in a pillar.
This is useful if the Pi is used upside down or sideways to correct the orientation of the image being shown.

Example:

.. code-block:: yaml

    sensehat:
        rotation: 90

"""

import logging

try:
    from sense_hat import SenseHat

    has_sense_hat = True
except (ImportError, NameError):
    _sensehat = None
    has_sense_hat = False

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load the module if SenseHat is available
    """
    if has_sense_hat:
        try:
            global _sensehat
            _sensehat = SenseHat()
        except OSError:
            return (
                False,
                "This module can only be used on a Raspberry Pi with a SenseHat.",
            )

        rotation = __salt__["pillar.get"]("sensehat:rotation", 0)
        if rotation in [0, 90, 180, 270]:
            _sensehat.set_rotation(rotation, False)
        else:
            log.error("%s is not a valid rotation. Using default rotation.", rotation)
        return True

    return (
        False,
        "The SenseHat execution module cannot be loaded: 'sense_hat' python library"
        " unavailable.",
    )


def set_pixels(pixels):
    """
    Sets the entire LED matrix based on a list of 64 pixel values

    pixels
        A list of 64 ``[R, G, B]`` color values.
    """
    _sensehat.set_pixels(pixels)
    return {"pixels": pixels}


def get_pixels():
    """
    Returns a list of 64 smaller lists of ``[R, G, B]`` pixels representing the
    the currently displayed image on the LED matrix.

    .. note::
        When using ``set_pixels`` the pixel values can sometimes change when
        you read them again using ``get_pixels``. This is because we specify each
        pixel element as 8 bit numbers (0 to 255) but when they're passed into the
        Linux frame buffer for the LED matrix the numbers are bit shifted down
        to fit into RGB 565. 5 bits for red, 6 bits for green and 5 bits for blue.
        The loss of binary precision when performing this conversion
        (3 bits lost for red, 2 for green and 3 for blue) accounts for the
        discrepancies you see.

        The ``get_pixels`` method provides an accurate representation of how the
        pixels end up in frame buffer memory after you have called ``set_pixels``.
    """
    return _sensehat.get_pixels()


def set_pixel(x, y, color):
    """
    Sets a single pixel on the LED matrix to a specified color.

    x
        The x coordinate of the pixel. Ranges from 0 on the left to 7 on the right.
    y
        The y coordinate of the pixel. Ranges from 0 at the top to 7 at the bottom.
    color
        The new color of the pixel as a list of ``[R, G, B]`` values.

    CLI Example:

    .. code-block:: bash

        salt 'raspberry' sensehat.set_pixel 0 0 '[255, 0, 0]'
    """
    _sensehat.set_pixel(x, y, color)
    return {"color": color}


def get_pixel(x, y):
    """
    Returns the color of a single pixel on the LED matrix.

    x
        The x coordinate of the pixel. Ranges from 0 on the left to 7 on the right.
    y
        The y coordinate of the pixel. Ranges from 0 at the top to 7 at the bottom.

    .. note::
        Please read the note for ``get_pixels``
    """
    return _sensehat.get_pixel(x, y)


def low_light(low_light=True):
    """
    Sets the LED matrix to low light mode. Useful in a dark environment.

    CLI Example:

    .. code-block:: bash

        salt 'raspberry' sensehat.low_light
        salt 'raspberry' sensehat.low_light False
    """
    _sensehat.low_light = low_light
    return {"low_light": low_light}


def show_message(
    message, msg_type=None, text_color=None, back_color=None, scroll_speed=0.1
):
    """
    Displays a message on the LED matrix.

    message
        The message to display
    msg_type
        The type of the message. Changes the appearance of the message.

        Available types are::

            error:      red text
            warning:    orange text
            success:    green text
            info:       blue text

    scroll_speed
        The speed at which the message moves over the LED matrix.
        This value represents the time paused for between shifting the text
        to the left by one column of pixels. Defaults to '0.1'.
    text_color
        The color in which the message is shown. Defaults to '[255, 255, 255]' (white).
    back_color
        The background color of the display. Defaults to '[0, 0, 0]' (black).

    CLI Example:

    .. code-block:: bash

        salt 'raspberry' sensehat.show_message 'Status ok'
        salt 'raspberry' sensehat.show_message 'Something went wrong' error
        salt 'raspberry' sensehat.show_message 'Red' text_color='[255, 0, 0]'
        salt 'raspberry' sensehat.show_message 'Hello world' None '[0, 0, 255]' '[255, 255, 0]' 0.2
    """
    text_color = text_color or [255, 255, 255]
    back_color = back_color or [0, 0, 0]

    color_by_type = {
        "error": [255, 0, 0],
        "warning": [255, 100, 0],
        "success": [0, 255, 0],
        "info": [0, 0, 255],
    }

    if msg_type in color_by_type:
        text_color = color_by_type[msg_type]

    _sensehat.show_message(message, scroll_speed, text_color, back_color)
    return {"message": message}


def show_letter(letter, text_color=None, back_color=None):
    """
    Displays a single letter on the LED matrix.

    letter
        The letter to display
    text_color
        The color in which the letter is shown. Defaults to '[255, 255, 255]' (white).
    back_color
        The background color of the display. Defaults to '[0, 0, 0]' (black).

    CLI Example:

    .. code-block:: bash

        salt 'raspberry' sensehat.show_letter O
        salt 'raspberry' sensehat.show_letter X '[255, 0, 0]'
        salt 'raspberry' sensehat.show_letter B '[0, 0, 255]' '[255, 255, 0]'
    """
    text_color = text_color or [255, 255, 255]
    back_color = back_color or [0, 0, 0]

    _sensehat.show_letter(letter, text_color, back_color)
    return {"letter": letter}


def show_image(image):
    """
    Displays an 8 x 8 image on the LED matrix.

    image
        The path to the image to display. The image must be 8 x 8 pixels in size.

    CLI Example:

    .. code-block:: bash

        salt 'raspberry' sensehat.show_image /tmp/my_image.png
    """
    return _sensehat.load_image(image)


def clear(color=None):
    """
    Sets the LED matrix to a single color or turns all LEDs off.

    CLI Example:

    .. code-block:: bash

        salt 'raspberry' sensehat.clear
        salt 'raspberry' sensehat.clear '[255, 0, 0]'
    """
    if color is None:
        _sensehat.clear()
    else:
        _sensehat.clear(color)
    return {"color": color}


def get_humidity():
    """
    Get the percentage of relative humidity from the humidity sensor.
    """
    return _sensehat.get_humidity()


def get_pressure():
    """
    Gets the current pressure in Millibars from the pressure sensor.
    """
    return _sensehat.get_pressure()


def get_temperature():
    """
    Gets the temperature in degrees Celsius from the humidity sensor.
    Equivalent to calling ``get_temperature_from_humidity``.

    If you get strange results try using ``get_temperature_from_pressure``.
    """
    return _sensehat.get_temperature()


def get_temperature_from_humidity():
    """
    Gets the temperature in degrees Celsius from the humidity sensor.
    """
    return _sensehat.get_temperature_from_humidity()


def get_temperature_from_pressure():
    """
    Gets the temperature in degrees Celsius from the pressure sensor.
    """
    return _sensehat.get_temperature_from_pressure()



import logging

import pytest
import pytestshellutils.utils.ports
import zmq
import zmq.eventloop.zmqstream

import salt.exceptions
import salt.ext.tornado.gen
import salt.ext.tornado.locks
import salt.transport.zeromq

log = logging.getLogger(__name__)


@pytest.fixture
def port():
    return pytestshellutils.utils.ports.get_unused_localhost_port()


@pytest.fixture
def request_client(io_loop, minion_opts, port):
    minion_opts["master_uri"] = f"tcp://127.0.0.1:{port}"
    client = salt.transport.zeromq.RequestClient(minion_opts, io_loop)
    try:
        yield client
    finally:
        client.close()


async def test_request_channel_issue_64627(io_loop, request_client, minion_opts, port):
    """
    Validate socket is preserved until request channel is explicitly closed.
    """
    minion_opts["master_uri"] = f"tcp://127.0.0.1:{port}"

    ctx = zmq.Context()
    socket = ctx.socket(zmq.REP)
    socket.bind(minion_opts["master_uri"])
    stream = zmq.eventloop.zmqstream.ZMQStream(socket, io_loop=io_loop)
    try:

        @salt.ext.tornado.gen.coroutine
        def req_handler(stream, msg):
            stream.send(msg[0])

        stream.on_recv_stream(req_handler)

        rep = await request_client.send(b"foo")
        req_socket = request_client.message_client.socket
        rep = await request_client.send(b"foo")
        assert req_socket is request_client.message_client.socket
        request_client.close()
        assert request_client.message_client.socket is None

    finally:
        stream.close()


async def test_request_channel_issue_65265(io_loop, request_client, minion_opts, port):

    minion_opts["master_uri"] = f"tcp://127.0.0.1:{port}"

    ctx = zmq.Context()
    socket = ctx.socket(zmq.REP)
    socket.bind(minion_opts["master_uri"])
    stream = zmq.eventloop.zmqstream.ZMQStream(socket, io_loop=io_loop)

    try:
        send_complete = salt.ext.tornado.locks.Event()

        @salt.ext.tornado.gen.coroutine
        def no_handler(stream, msg):
            """
            The server never responds.
            """
            stream.close()

        stream.on_recv_stream(no_handler)

        @salt.ext.tornado.gen.coroutine
        def send_request():
            """
            The request will timeout becuse the server does not respond.
            """
            ret = None
            with pytest.raises(salt.exceptions.SaltReqTimeoutError):
                yield request_client.send("foo", timeout=3)
            send_complete.set()
            return ret

        io_loop.spawn_callback(send_request)

        await send_complete.wait()

    finally:
        stream.close()

    # Create a new server, the old socket has been closed.

    @salt.ext.tornado.gen.coroutine
    def req_handler(stream, msg):
        """
        The server responds
        """
        stream.send(salt.payload.dumps("bar"))

    socket = ctx.socket(zmq.REP)
    socket.bind(minion_opts["master_uri"])
    stream = zmq.eventloop.zmqstream.ZMQStream(socket, io_loop=io_loop)
    try:
        stream.on_recv_stream(req_handler)
        await salt.ext.tornado.gen.sleep(1)

        ret = await request_client.send("foo", timeout=1)
        assert ret == "bar"
    finally:
        stream.close()


async def test_request_client_send_recv_socket_closed(
    io_loop, request_client, minion_opts, port, caplog
):
    minion_opts["master_uri"] = f"tcp://127.0.0.1:{port}"
    ctx = zmq.Context()
    socket = ctx.socket(zmq.REP)
    socket.bind(minion_opts["master_uri"])
    stream = zmq.eventloop.zmqstream.ZMQStream(socket, io_loop=io_loop)

    request_client.connect()
    socket = request_client.message_client.socket
    with caplog.at_level(logging.TRACE):
        request_client.close()
        await salt.ext.tornado.gen.sleep(0.5)
        assert "Send socket closed while polling." in caplog.messages
        assert f"Send and receive coroutine ending {socket}" in caplog.messages


def test_request_client_send_recv_loop_closed(minion_opts, port, caplog):
    io_loop = salt.ext.tornado.ioloop.IOLoop()
    minion_opts["master_uri"] = f"tcp://127.0.0.1:{port}"
    ctx = zmq.Context()
    serve_socket = ctx.socket(zmq.REP)
    serve_socket.bind(minion_opts["master_uri"])
    minion_opts["master_uri"] = f"tcp://127.0.0.1:{port}"
    request_client = salt.transport.zeromq.RequestClient(minion_opts, io_loop)

    request_client.connect()

    def poll(*args, **kwargs):
        """
        Mock this error because it is incredibly hard to time this.
        """
        raise zmq.eventloop.future.CancelledError()

    socket = request_client.message_client.socket
    socket.poll = poll

    with caplog.at_level(logging.TRACE):

        @salt.ext.tornado.gen.coroutine
        def testit():
            yield salt.ext.tornado.gen.sleep(0.5)
            io_loop.stop()

        io_loop.add_callback(testit)
        io_loop.start()

        try:
            assert "Loop closed while polling send socket." in caplog.messages
            assert f"Send and receive coroutine ending {socket}" in caplog.messages
        finally:
            request_client.close()
            serve_socket.close()


@pytest.mark.parametrize(
    "errno", [zmq.ETERM, zmq.ENOTSOCK, zmq.error.EINTR, zmq.EFSM, 321]
)
async def test_request_client_send_msg_socket_closed(
    io_loop, request_client, minion_opts, port, caplog, errno
):
    minion_opts["master_uri"] = f"tcp://127.0.0.1:{port}"
    ctx = zmq.Context()
    serve_socket = ctx.socket(zmq.REP)
    serve_socket.bind(minion_opts["master_uri"])

    request_client.connect()

    @salt.ext.tornado.gen.coroutine
    def send(*args, **kwargs):
        """
        Mock this error because it is incredibly hard to time this.
        """
        raise zmq.ZMQError(errno=errno)

    socket = request_client.message_client.socket
    socket.send = send
    with caplog.at_level(logging.TRACE):
        with pytest.raises(zmq.ZMQError):
            try:
                await request_client.send("meh")
                await salt.ext.tornado.gen.sleep(0.3)
                if errno == zmq.EFSM:
                    assert "Socket was found in invalid state." in caplog.messages
                elif errno != 321:
                    assert "Recieve socket closed while polling." in caplog.messages
                else:
                    assert (
                        "Unhandled Zeromq error durring send/receive: Unknown error 321"
                        in caplog.messages
                    )
                assert (
                    "The request timed out or ended with an error before sending completed. reconnecting."
                    in caplog.messages
                )
                assert f"Send and receive coroutine ending {socket}" in caplog.messages
            finally:
                request_client.close()
                serve_socket.close()


async def test_request_client_send_msg_loop_closed(
    io_loop, request_client, minion_opts, port, caplog
):
    minion_opts["master_uri"] = f"tcp://127.0.0.1:{port}"
    ctx = zmq.Context()
    serve_socket = ctx.socket(zmq.REP)
    serve_socket.bind(minion_opts["master_uri"])

    request_client.connect()

    @salt.ext.tornado.gen.coroutine
    def send(*args, **kwargs):
        """
        Mock this error because it is incredibly hard to time this.
        """
        raise zmq.eventloop.future.CancelledError()

    socket = request_client.message_client.socket
    socket.send = send
    with caplog.at_level(logging.TRACE):
        with pytest.raises(zmq.eventloop.future.CancelledError):
            try:
                await request_client.send("meh")
                await salt.ext.tornado.gen.sleep(0.3)
                assert "Loop closed while sending." in caplog.messages
                assert f"Send and receive coroutine ending {socket}" in caplog.messages
            finally:
                request_client.close()
                serve_socket.close()


async def test_request_client_recv_poll_loop_closed(
    io_loop, request_client, minion_opts, port, caplog
):
    minion_opts["master_uri"] = f"tcp://127.0.0.1:{port}"
    ctx = zmq.Context()
    serve_socket = ctx.socket(zmq.REP)
    serve_socket.bind(minion_opts["master_uri"])

    request_client.connect()

    socket = request_client.message_client.socket

    def poll(*args, **kwargs):
        """
        Mock this error because it is incredibly hard to time this.
        """
        if args[1] == zmq.POLLIN:
            raise zmq.eventloop.future.CancelledError()
        else:
            return socket.poll(*args, **kwargs)

    socket.poll = poll
    with caplog.at_level(logging.TRACE):
        with pytest.raises(zmq.eventloop.future.CancelledError):
            try:
                await request_client.send("meh")
                await salt.ext.tornado.gen.sleep(0.3)
                assert "Loop closed while polling receive socket." in caplog.messages
                assert f"Send and receive coroutine ending {socket}" in caplog.messages
            finally:
                request_client.close()
                serve_socket.close()


async def test_request_client_recv_poll_socket_closed(
    io_loop, request_client, minion_opts, port, caplog
):
    minion_opts["master_uri"] = f"tcp://127.0.0.1:{port}"
    ctx = zmq.Context()
    serve_socket = ctx.socket(zmq.REP)
    serve_socket.bind(minion_opts["master_uri"])

    request_client.connect()

    socket = request_client.message_client.socket

    def poll(*args, **kwargs):
        """
        Mock this error because it is incredibly hard to time this.
        """
        if args[1] == zmq.POLLIN:
            raise zmq.ZMQError()
        else:
            return socket.poll(*args, **kwargs)

    socket.poll = poll
    with caplog.at_level(logging.TRACE):
        with pytest.raises(zmq.ZMQError):
            try:
                await request_client.send("meh")
                await salt.ext.tornado.gen.sleep(0.3)
                assert "Recieve socket closed while polling." in caplog.messages
                assert f"Send and receive coroutine ending {socket}" in caplog.messages
            finally:
                request_client.close()
                serve_socket.close()


async def test_request_client_recv_loop_closed(
    io_loop, request_client, minion_opts, port, caplog
):
    minion_opts["master_uri"] = f"tcp://127.0.0.1:{port}"
    ctx = zmq.Context()
    serve_socket = ctx.socket(zmq.REP)
    serve_socket.bind(minion_opts["master_uri"])
    stream = zmq.eventloop.zmqstream.ZMQStream(serve_socket, io_loop=io_loop)

    @salt.ext.tornado.gen.coroutine
    def req_handler(stream, msg):
        stream.send(msg[0])

    stream.on_recv_stream(req_handler)

    request_client.connect()

    socket = request_client.message_client.socket

    @salt.ext.tornado.gen.coroutine
    def recv(*args, **kwargs):
        """
        Mock this error because it is incredibly hard to time this.
        """
        raise zmq.eventloop.future.CancelledError()

    socket.recv = recv

    with caplog.at_level(logging.TRACE):
        with pytest.raises(zmq.eventloop.future.CancelledError):
            try:
                await request_client.send("meh")
                await salt.ext.tornado.gen.sleep(0.3)
                assert "Loop closed while receiving." in caplog.messages
                assert f"Send and receive coroutine ending {socket}" in caplog.messages
            finally:
                request_client.close()
                serve_socket.close()
                ctx.term()


async def test_request_client_recv_socket_closed(
    io_loop, request_client, minion_opts, port, caplog
):
    minion_opts["master_uri"] = f"tcp://127.0.0.1:{port}"
    ctx = zmq.Context()
    serve_socket = ctx.socket(zmq.REP)
    serve_socket.bind(minion_opts["master_uri"])
    stream = zmq.eventloop.zmqstream.ZMQStream(serve_socket, io_loop=io_loop)

    @salt.ext.tornado.gen.coroutine
    def req_handler(stream, msg):
        stream.send(msg[0])

    stream.on_recv_stream(req_handler)

    request_client.connect()

    socket = request_client.message_client.socket

    @salt.ext.tornado.gen.coroutine
    def recv(*args, **kwargs):
        """
        Mock this error because it is incredibly hard to time this.
        """
        raise zmq.ZMQError()

    socket.recv = recv

    with caplog.at_level(logging.TRACE):
        with pytest.raises(zmq.ZMQError):
            try:
                await request_client.send("meh")
                await salt.ext.tornado.gen.sleep(0.3)
                assert "Receive socket closed while receiving." in caplog.messages
                assert f"Send and receive coroutine ending {socket}" in caplog.messages
            finally:
                request_client.close()
                serve_socket.close()
                ctx.term()



"""
This thorium state is used to track the status beacon events and keep track of
the active status of minions

.. versionadded:: 2016.11.0
"""

import fnmatch
import time


def reg(name):
    """
    Activate this register to turn on a minion status tracking register, this
    register keeps the current status beacon data and the time that each beacon
    was last checked in.
    """
    ret = {"name": name, "changes": {}, "comment": "", "result": True}
    now = time.time()
    if "status" not in __reg__:
        __reg__["status"] = {}
        __reg__["status"]["val"] = {}
    for event in __events__:
        if fnmatch.fnmatch(event["tag"], "salt/beacon/*/status/*"):
            # Got one!
            idata = {"recv_time": now}
            for key in event["data"]["data"]:
                if key in ("id", "recv_time"):
                    continue
                idata[key] = event["data"]["data"][key]
            __reg__["status"]["val"][event["data"]["id"]] = idata
            ret["changes"][event["data"]["id"]] = True
    return ret



"""
Module for execution of ServiceNow CI (configuration items)

.. versionadded:: 2016.11.0

:depends: servicenow_rest python module

:configuration: Configure this module by specifying the name of a configuration
    profile in the minion config, minion pillar, or master config. The module
    will use the 'servicenow' key by default, if defined.

    For example:

    .. code-block:: yaml

        servicenow:
          instance_name: ''
          username: ''
          password: ''
"""

import logging

HAS_LIBS = False
try:
    from servicenow_rest.api import Client

    HAS_LIBS = True
except ImportError:
    pass

log = logging.getLogger(__name__)

__virtualname__ = "servicenow"

SERVICE_NAME = "servicenow"


def __virtual__():
    """
    Only load this module if servicenow is installed on this minion.
    """
    if HAS_LIBS:
        return __virtualname__
    return (
        False,
        "The servicenow execution module failed to load: "
        "requires servicenow_rest python library to be installed.",
    )


def _get_client():
    config = __salt__["config.option"](SERVICE_NAME)
    instance_name = config["instance_name"]
    username = config["username"]
    password = config["password"]
    return Client(instance_name, username, password)


def set_change_request_state(change_id, state="approved"):
    """
    Set the approval state of a change request/record

    :param change_id: The ID of the change request, e.g. CHG123545
    :type  change_id: ``str``

    :param state: The target state, e.g. approved
    :type  state: ``str``

    CLI Example:

    .. code-block:: bash

        salt myminion servicenow.set_change_request_state CHG000123 declined
        salt myminion servicenow.set_change_request_state CHG000123 approved
    """
    client = _get_client()
    client.table = "change_request"
    # Get the change record first
    record = client.get({"number": change_id})
    if not record:
        log.error("Failed to fetch change record, maybe it does not exist?")
        return False
    # Use the sys_id as the unique system record
    sys_id = record[0]["sys_id"]
    response = client.update({"approval": state}, sys_id)
    return response


def delete_record(table, sys_id):
    """
    Delete an existing record

    :param table: The table name, e.g. sys_user
    :type  table: ``str``

    :param sys_id: The unique ID of the record
    :type  sys_id: ``str``

    CLI Example:

    .. code-block:: bash

        salt myminion servicenow.delete_record sys_computer 2134566
    """
    client = _get_client()
    client.table = table
    response = client.delete(sys_id)
    return response


def non_structured_query(table, query=None, **kwargs):
    """
    Run a non-structed (not a dict) query on a servicenow table.
    See http://wiki.servicenow.com/index.php?title=Encoded_Query_Strings#gsc.tab=0
    for help on constructing a non-structured query string.

    :param table: The table name, e.g. sys_user
    :type  table: ``str``

    :param query: The query to run (or use keyword arguments to filter data)
    :type  query: ``str``

    CLI Example:

    .. code-block:: bash

        salt myminion servicenow.non_structured_query sys_computer 'role=web'
        salt myminion servicenow.non_structured_query sys_computer role=web type=computer
    """
    client = _get_client()
    client.table = table
    # underlying lib doesn't use six or past.basestring,
    # does isinstance(x, str)
    # http://bit.ly/1VkMmpE
    if query is None:
        # try and assemble a query by keyword
        query_parts = []
        for key, value in kwargs.items():
            query_parts.append(f"{key}={value}")
        query = "^".join(query_parts)
    query = str(query)
    response = client.get(query)
    return response


def update_record_field(table, sys_id, field, value):
    """
    Update the value of a record's field in a servicenow table

    :param table: The table name, e.g. sys_user
    :type  table: ``str``

    :param sys_id: The unique ID of the record
    :type  sys_id: ``str``

    :param field: The new value
    :type  field: ``str``

    :param value: The new value
    :type  value: ``str``

    CLI Example:

    .. code-block:: bash

        salt myminion servicenow.update_record_field sys_user 2348234 first_name jimmy
    """
    client = _get_client()
    client.table = table
    response = client.update({field: value}, sys_id)
    return response



#!/usr/bin/env python
"""
Test interacting with the wheel system. This script is useful when testing
wheel modules
"""
import optparse  # pylint: disable=deprecated-module
import pprint

import salt.auth
import salt.config
import salt.wheel


def parse():
    """
    Parse the command line options
    """
    parser = optparse.OptionParser()
    parser.add_option(
        "-f", "--fun", "--function", dest="fun", help="The wheel function to execute"
    )
    parser.add_option(
        "-a",
        "--auth",
        dest="eauth",
        help="The external authentication mechanism to use",
    )

    options, args = parser.parse_args()

    cli = options.__dict__

    for arg in args:
        if "=" in arg:
            comps = arg.split("=")
            cli[comps[0]] = comps[1]
    return cli


class Wheeler:
    """
    Set up communication with the wheel interface
    """

    def __init__(self, cli):
        self.opts = salt.config.master_config("/etc/salt")
        self.opts.update(cli)
        self.__eauth()
        self.wheel = salt.wheel.Wheel(self.opts)

    def __eauth(self):
        """
        Fill in the blanks for the eauth system
        """
        if self.opts["eauth"]:
            resolver = salt.auth.Resolver(self.opts)
            res = resolver.cli(self.opts["eauth"])
        self.opts.update(res)

    def run(self):
        """
        Execute the wheel call
        """
        return self.wheel.master_call(**self.opts)


if __name__ == "__main__":
    wheeler = Wheeler(parse())
    pprint.pprint(wheeler.run())



"""
Packet Cloud Module Using Packet's Python API Client
====================================================

The Packet cloud module is used to control access to the Packet VPS system.

Use of this module only requires the ``token`` parameter.

Set up the cloud configuration at ``/etc/salt/cloud.providers`` or ``/etc/salt/cloud.providers.d/packet.conf``:

The Packet profile requires ``size``, ``image``, ``location``,  ``project_id``

Optional profile parameters:

- ``storage_size`` -  min value is 10, defines Gigabytes of storage that will be attached to device.
- ``storage_tier`` - storage_1 - Standard Plan, storage_2 - Performance Plan
- ``snapshot_count`` - int
- ``snapshot_frequency`` - string - possible values:

  - 1min
  - 15min
  - 1hour
  - 1day
  - 1week
  - 1month
  - 1year

This driver requires Packet's client library: https://pypi.python.org/pypi/packet-python

.. code-block:: yaml

    packet-provider:
        minion:
            master: 192.168.50.10
        driver: packet
        token: ewr23rdf35wC8oNjJrhmHa87rjSXzJyi
        private_key: /root/.ssh/id_rsa

    packet-profile:
        provider: packet-provider
        size: baremetal_0
        image: ubuntu_16_04_image
        location: ewr1
        project_id: a64d000b-d47c-4d26-9870-46aac43010a6
        storage_size: 10
        storage_tier: storage_1
        storage_snapshot_count: 1
        storage_snapshot_frequency: 15min
"""

import logging
import pprint
import time

import salt.config as config
import salt.utils.cloud
from salt.cloud.libcloudfuncs import get_image, get_size, script, show_instance
from salt.exceptions import SaltCloudException, SaltCloudSystemExit
from salt.utils.functools import namespaced_function

try:
    import packet

    HAS_PACKET = True
except ImportError:
    HAS_PACKET = False


get_size = namespaced_function(get_size, globals())
get_image = namespaced_function(get_image, globals())

script = namespaced_function(script, globals())

show_instance = namespaced_function(show_instance, globals())

# Get logging started
log = logging.getLogger(__name__)

__virtualname__ = "packet"


# Only load this module if the Packet configuration is in place.
def __virtual__():
    """
    Check for Packet configs.
    """
    if HAS_PACKET is False:
        return False, "The packet python library is not installed"
    if get_configured_provider() is False:
        return False

    return __virtualname__


def _get_active_provider_name():
    try:
        return __active_provider_name__.value()
    except AttributeError:
        return __active_provider_name__


def get_configured_provider():
    """
    Return the first configured instance.
    """
    return config.is_provider_configured(
        __opts__, _get_active_provider_name() or __virtualname__, ("token",)
    )


def avail_images(call=None):
    """
    Return available Packet os images.

    CLI Example:

    .. code-block:: bash

        salt-cloud --list-images packet-provider
        salt-cloud -f avail_images packet-provider
    """
    if call == "action":
        raise SaltCloudException(
            "The avail_images function must be called with -f or --function."
        )

    ret = {}

    vm_ = get_configured_provider()
    manager = packet.Manager(auth_token=vm_["token"])

    ret = {}

    for os_system in manager.list_operating_systems():
        ret[os_system.name] = os_system.__dict__

    return ret


def avail_locations(call=None):
    """
    Return available Packet datacenter locations.

    CLI Example:

    .. code-block:: bash

        salt-cloud --list-locations packet-provider
        salt-cloud -f avail_locations packet-provider
    """
    if call == "action":
        raise SaltCloudException(
            "The avail_locations function must be called with -f or --function."
        )

    vm_ = get_configured_provider()
    manager = packet.Manager(auth_token=vm_["token"])

    ret = {}

    for facility in manager.list_facilities():
        ret[facility.name] = facility.__dict__

    return ret


def avail_sizes(call=None):
    """
    Return available Packet sizes.

    CLI Example:

    .. code-block:: bash

        salt-cloud --list-sizes packet-provider
        salt-cloud -f avail_sizes packet-provider
    """
    if call == "action":
        raise SaltCloudException(
            "The avail_locations function must be called with -f or --function."
        )

    vm_ = get_configured_provider()

    manager = packet.Manager(auth_token=vm_["token"])

    ret = {}

    for plan in manager.list_plans():
        ret[plan.name] = plan.__dict__

    return ret


def avail_projects(call=None):
    """
    Return available Packet projects.

    CLI Example:

    .. code-block:: bash

        salt-cloud -f avail_projects packet-provider
    """
    if call == "action":
        raise SaltCloudException(
            "The avail_projects function must be called with -f or --function."
        )

    vm_ = get_configured_provider()
    manager = packet.Manager(auth_token=vm_["token"])

    ret = {}

    for project in manager.list_projects():
        ret[project.name] = project.__dict__

    return ret


def _wait_for_status(status_type, object_id, status=None, timeout=500, quiet=True):
    """
    Wait for a certain status from Packet.
    status_type
        device or volume
    object_id
        The ID of the Packet device or volume to wait on. Required.
    status
        The status to wait for.
    timeout
        The amount of time to wait for a status to update.
    quiet
        Log status updates to debug logs when False. Otherwise, logs to info.
    """
    if status is None:
        status = "ok"

    interval = 5
    iterations = int(timeout / interval)

    vm_ = get_configured_provider()
    manager = packet.Manager(auth_token=vm_["token"])

    for i in range(0, iterations):
        get_object = getattr(manager, f"get_{status_type}")
        obj = get_object(object_id)

        if obj.state == status:
            return obj

        time.sleep(interval)
        log.log(
            logging.INFO if not quiet else logging.DEBUG,
            "Status for Packet %s is '%s', waiting for '%s'.",
            object_id,
            obj.state,
            status,
        )

    return obj


def is_profile_configured(vm_):
    try:
        # Check for required profile parameters before sending any API calls.
        if (
            vm_["profile"]
            and config.is_profile_configured(
                __opts__,
                _get_active_provider_name() or __virtualname__,
                vm_["profile"],
                vm_=vm_,
            )
            is False
        ):
            return False

        alias, driver = _get_active_provider_name().split(":")

        profile_data = __opts__["providers"][alias][driver]["profiles"][vm_["profile"]]

        if profile_data.get("storage_size") or profile_data.get("storage_tier"):
            required_keys = ["storage_size", "storage_tier"]

            for key in required_keys:
                if profile_data.get(key) is None:
                    log.error(
                        "both storage_size and storage_tier required for "
                        "profile %s. Please check your profile configuration",
                        vm_["profile"],
                    )
                    return False

            locations = avail_locations()

            for location in locations.values():
                if location["code"] == profile_data["location"]:
                    if "storage" not in location["features"]:
                        log.error(
                            "Chosen location %s for profile %s does not "
                            "support storage feature. Please check your "
                            "profile configuration",
                            location["code"],
                            vm_["profile"],
                        )
                        return False

        if profile_data.get("storage_snapshot_count") or profile_data.get(
            "storage_snapshot_frequency"
        ):
            required_keys = ["storage_size", "storage_tier"]

            for key in required_keys:
                if profile_data.get(key) is None:
                    log.error(
                        "both storage_snapshot_count and "
                        "storage_snapshot_frequency required for profile "
                        "%s. Please check your profile configuration",
                        vm_["profile"],
                    )
                    return False

    except AttributeError:
        pass

    return True


def create(vm_):
    """
    Create a single Packet VM.
    """
    name = vm_["name"]

    if not is_profile_configured(vm_):
        return False

    __utils__["cloud.fire_event"](
        "event",
        "starting create",
        f"salt/cloud/{name}/creating",
        args=__utils__["cloud.filter_event"](
            "creating", vm_, ["name", "profile", "provider", "driver"]
        ),
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    log.info("Creating Packet VM %s", name)

    manager = packet.Manager(auth_token=vm_["token"])

    __utils__["cloud.fire_event"](
        "event",
        "requesting instance",
        "salt/cloud/{}/requesting".format(vm_["name"]),
        args=__utils__["cloud.filter_event"](
            "requesting", vm_, ["name", "profile", "provider", "driver"]
        ),
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    device = manager.create_device(
        project_id=vm_["project_id"],
        hostname=name,
        plan=vm_["size"],
        facility=vm_["location"],
        operating_system=vm_["image"],
    )

    device = _wait_for_status("device", device.id, status="active")

    if device.state != "active":
        log.error(
            "Error creating %s on PACKET\n\nwhile waiting for initial ready status",
            name,
            exc_info_on_loglevel=logging.DEBUG,
        )

    # Define which ssh_interface to use
    ssh_interface = _get_ssh_interface(vm_)

    # Pass the correct IP address to the bootstrap ssh_host key
    if ssh_interface == "private_ips":
        for ip in device.ip_addresses:
            if ip["public"] is False:
                vm_["ssh_host"] = ip["address"]
                break
    else:
        for ip in device.ip_addresses:
            if ip["public"] is True:
                vm_["ssh_host"] = ip["address"]
                break

    key_filename = config.get_cloud_config_value(
        "private_key", vm_, __opts__, search_global=False, default=None
    )

    vm_["key_filename"] = key_filename

    vm_["private_key"] = key_filename

    # Bootstrap!
    ret = __utils__["cloud.bootstrap"](vm_, __opts__)

    ret.update({"device": device.__dict__})

    if vm_.get("storage_tier") and vm_.get("storage_size"):
        # create storage and attach it to device

        volume = manager.create_volume(
            vm_["project_id"],
            f"{name}_storage",
            vm_.get("storage_tier"),
            vm_.get("storage_size"),
            vm_.get("location"),
            snapshot_count=vm_.get("storage_snapshot_count", 0),
            snapshot_frequency=vm_.get("storage_snapshot_frequency"),
        )

        volume.attach(device.id)

        volume = _wait_for_status("volume", volume.id, status="active")

        if volume.state != "active":
            log.error(
                "Error creating %s on PACKET\n\nwhile waiting for initial ready status",
                name,
                exc_info_on_loglevel=logging.DEBUG,
            )

        ret.update({"volume": volume.__dict__})

    log.info("Created Cloud VM '%s'", name)

    log.debug("'%s' VM creation details:\n%s", name, pprint.pformat(device.__dict__))

    __utils__["cloud.fire_event"](
        "event",
        "created instance",
        f"salt/cloud/{name}/created",
        args=__utils__["cloud.filter_event"](
            "created", vm_, ["name", "profile", "provider", "driver"]
        ),
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    return ret


def list_nodes_full(call=None):
    """
    List devices, with all available information.

    CLI Example:

    .. code-block:: bash

        salt-cloud -F
        salt-cloud --full-query
        salt-cloud -f list_nodes_full packet-provider

    ..
    """
    if call == "action":
        raise SaltCloudException(
            "The list_nodes_full function must be called with -f or --function."
        )

    ret = {}

    for device in get_devices_by_token():
        ret[device.hostname] = device.__dict__

    return ret


def list_nodes_min(call=None):
    """
    Return a list of the VMs that are on the provider. Only a list of VM names and
    their state is returned. This is the minimum amount of information needed to
    check for existing VMs.

    .. versionadded:: 2015.8.0

    CLI Example:

    .. code-block:: bash

        salt-cloud -f list_nodes_min packet-provider
        salt-cloud --function list_nodes_min packet-provider
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_nodes_min function must be called with -f or --function."
        )

    ret = {}

    for device in get_devices_by_token():
        ret[device.hostname] = {"id": device.id, "state": device.state}

    return ret


def list_nodes_select(call=None):
    """
    Return a list of the VMs that are on the provider, with select fields.
    """
    return salt.utils.cloud.list_nodes_select(
        list_nodes_full(),
        __opts__["query.selection"],
        call,
    )


def get_devices_by_token():
    vm_ = get_configured_provider()
    manager = packet.Manager(auth_token=vm_["token"])

    devices = []

    for profile_name in vm_["profiles"]:
        profile = vm_["profiles"][profile_name]

        devices.extend(manager.list_devices(profile["project_id"]))

    return devices


def list_nodes(call=None):
    """
    Returns a list of devices, keeping only a brief listing.

    CLI Example:

    .. code-block:: bash

        salt-cloud -Q
        salt-cloud --query
        salt-cloud -f list_nodes packet-provider
    ..
    """

    if call == "action":
        raise SaltCloudException(
            "The list_nodes function must be called with -f or --function."
        )

    ret = {}

    for device in get_devices_by_token():
        ret[device.hostname] = device.__dict__

    return ret


def destroy(name, call=None):
    """
    Destroys a Packet device by name.

    name
        The hostname of VM to be be destroyed.

    CLI Example:

    .. code-block:: bash

        salt-cloud -d name
    """
    if call == "function":
        raise SaltCloudException(
            "The destroy action must be called with -d, --destroy, -a or --action."
        )

    __utils__["cloud.fire_event"](
        "event",
        "destroying instance",
        f"salt/cloud/{name}/destroying",
        args={"name": name},
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    vm_ = get_configured_provider()
    manager = packet.Manager(auth_token=vm_["token"])

    nodes = list_nodes_min()

    node = nodes[name]

    for project in manager.list_projects():

        for volume in manager.list_volumes(project.id):
            if volume.attached_to == node["id"]:
                volume.detach()
                volume.delete()
                break

    manager.call_api("devices/{id}".format(id=node["id"]), type="DELETE")

    __utils__["cloud.fire_event"](
        "event",
        "destroyed instance",
        f"salt/cloud/{name}/destroyed",
        args={"name": name},
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    return {}


def _get_ssh_interface(vm_):
    """
    Return the ssh_interface type to connect to. Either 'public_ips' (default)
    or 'private_ips'.
    """
    return config.get_cloud_config_value(
        "ssh_interface", vm_, __opts__, default="public_ips", search_global=False
    )



import contextlib
import logging

import pytest

import salt.client.ssh.wrapper.slsutil as slsutil
from salt.exceptions import CommandExecutionError, SaltInvocationError
from tests.support.mock import MagicMock

log = logging.getLogger(__name__)


# --- These tests are adapted from tests.pytests.unit.utils.slsutil


@pytest.fixture
def configure_loader_modules(master_dirs, master_files):
    return {
        slsutil: {
            "__salt__": {
                "cp.list_master": MagicMock(return_value=master_files),
                "cp.list_master_dirs": MagicMock(return_value=master_dirs),
            },
            "__opts__": {
                "renderer": "jinja|yaml",
                "renderer_blacklist": [],
                "renderer_whitelist": [],
            },
        }
    }


@pytest.fixture
def master_dirs():
    return ["red", "red/files", "blue", "blue/files"]


@pytest.fixture
def master_files():
    return [
        "top.sls",
        "red/init.sls",
        "red/files/default.conf",
        "blue/init.sls",
        "blue/files/default.conf",
    ]


@pytest.mark.parametrize("inpt,expected", ((True, "yes"), (False, "no")))
def test_boolstr(inpt, expected):
    assert slsutil.boolstr(inpt, true="yes", false="no") == expected


@pytest.mark.parametrize(
    "inpt,expected", (("red/init.sls", True), ("green/init.sls", False))
)
def test_file_exists(inpt, expected):
    assert slsutil.file_exists(inpt) is expected


@pytest.mark.parametrize("inpt,expected", (("red", True), ("green", False)))
def test_dir_exists(inpt, expected):
    assert slsutil.dir_exists(inpt) is expected


@pytest.mark.parametrize(
    "inpt,expected",
    (
        ("red", True),
        ("green", False),
        ("red/init.sls", True),
        ("green/init.sls", False),
    ),
)
def test_path_exists(inpt, expected):
    assert slsutil.path_exists(inpt) is expected


@pytest.mark.parametrize(
    "inpt,expected,raises",
    [
        (("red/files", "init.sls"), "red/init.sls", None),
        (("red/files", ["top.sls"]), "top.sls", None),
        (("", "top.sls"), "top.sls", None),
        ((None, "top.sls"), "top.sls", None),
        (("red/files", ["top.sls", "init.sls"]), "red/init.sls", None),
        (
            ("red/files", "notfound"),
            None,
            pytest.raises(
                CommandExecutionError, match=r"File pattern\(s\) not found.*"
            ),
        ),
        (
            ("red", "default.conf"),
            None,
            pytest.raises(
                CommandExecutionError, match=r"File pattern\(s\) not found.*"
            ),
        ),
        (
            ("green", "notfound"),
            None,
            pytest.raises(SaltInvocationError, match="Starting path not found.*"),
        ),
        (
            ("red", 1234),
            None,
            pytest.raises(
                SaltInvocationError, match=".*must be a string or list of strings.*"
            ),
        ),
    ],
)
def test_findup(inpt, expected, raises):
    if raises is None:
        raises = contextlib.nullcontext()
    with raises:
        res = slsutil.findup(*inpt)
        assert res == expected


@pytest.mark.parametrize(
    "a,b,merge_lists,expected",
    [
        (
            {"foo": {"bar": "baz", "hi": "there", "some": ["list"]}},
            {"foo": {"baz": "quux", "bar": "hi", "some": ["other_list"]}},
            False,
            {
                "foo": {
                    "baz": "quux",
                    "bar": "hi",
                    "hi": "there",
                    "some": ["other_list"],
                }
            },
        ),
        (
            {"foo": {"bar": "baz", "hi": "there", "some": ["list"]}},
            {"foo": {"baz": "quux", "bar": "hi", "some": ["other_list"]}},
            True,
            {
                "foo": {
                    "baz": "quux",
                    "bar": "hi",
                    "hi": "there",
                    "some": ["list", "other_list"],
                }
            },
        ),
    ],
)
@pytest.mark.parametrize("func", ("update", "merge", "merge_all"))
def test_update_merge(a, b, merge_lists, expected, func):
    arg = (a, b)
    if func == "merge_all":
        arg = ([a, b],)
    res = getattr(slsutil, func)(*arg, merge_lists=merge_lists)
    assert res == expected
    assert (a is res) is (func == "update")


def test_renderer_requires_either_path_or_string():
    with pytest.raises(SaltInvocationError, match=".*either path or string.*"):
        slsutil.renderer()



"""
Management of logs using Solaris logadm.

:maintainer:    Jorge Schrauwen <sjorge@blackdot.be>
:maturity:      new
:depends:       salt.modulus.logadm
:platform:      Oracle Solaris, Sun Solaris, illumos

.. versionadded:: 2017.7.0

.. code-block:: yaml

    .. note::
        TODO

"""

import logging

import salt.utils.args
import salt.utils.data

log = logging.getLogger(__name__)

# Define the state's virtual name
__virtualname__ = "logadm"


def __virtual__():
    """
    Provides logadm state if we have the module
    """
    if "logadm.list_conf" in __salt__:
        return True
    return (
        False,
        "{} state module can only if the logadm execution module is present".format(
            __virtualname__
        ),
    )


def rotate(name, **kwargs):
    """
    Add a log to the logadm configuration

    name : string
        alias for entryname

    kwargs : boolean|string|int
        optional additional flags and parameters

    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    # cleanup kwargs
    kwargs = salt.utils.args.clean_kwargs(**kwargs)

    # inject name as entryname
    if "entryname" not in kwargs:
        kwargs["entryname"] = name

    # figure out log_file and entryname
    if "log_file" not in kwargs or not kwargs["log_file"]:
        if "entryname" in kwargs and kwargs["entryname"]:
            if kwargs["entryname"].startswith("/"):
                kwargs["log_file"] = kwargs["entryname"]

    # check for log_file
    if "log_file" not in kwargs or not kwargs["log_file"]:
        ret["result"] = False
        ret["comment"] = "Missing log_file attribute!"
    else:
        # lookup old configuration
        old_config = __salt__["logadm.list_conf"]()

        # remove existing entry
        if kwargs["log_file"] in old_config:
            res = __salt__["logadm.remove"](
                kwargs["entryname"] if "entryname" in kwargs else kwargs["log_file"]
            )
            ret["result"] = "Error" not in res
            if not ret["result"]:
                ret["comment"] = res["Error"]
                ret["changes"] = {}

        # add new entry
        res = __salt__["logadm.rotate"](name, **kwargs)
        ret["result"] = "Error" not in res
        if ret["result"]:
            new_config = __salt__["logadm.list_conf"]()
            ret["comment"] = "Log configuration {}".format(
                "updated" if kwargs["log_file"] in old_config else "added"
            )
            if kwargs["log_file"] in old_config:
                for key, val in salt.utils.data.compare_dicts(
                    old_config[kwargs["log_file"]], new_config[kwargs["log_file"]]
                ).items():
                    ret["changes"][key] = val["new"]
            else:
                ret["changes"] = new_config[kwargs["log_file"]]
            log.debug(ret["changes"])
        else:
            ret["comment"] = res["Error"]
            # NOTE: we need to remove the log file first
            #       potentially the log configuraiton can get lost :s
            if kwargs["log_file"] in old_config:
                ret["changes"] = {kwargs["log_file"]: None}
            else:
                ret["changes"] = {}

    return ret


def remove(name, log_file=None):
    """
    Remove a log from the logadm configuration

    name : string
        entryname
    log_file : string
        (optional) log file path

    .. note::
        If log_file is specified it will be used instead of the entry name.

    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    # retrieve all log configuration
    config = __salt__["logadm.list_conf"]()

    # figure out log_file and name
    if not log_file:
        if name.startswith("/"):
            log_file = name
            name = None
        else:
            for log in config:
                if "entryname" in config[log] and config[log]["entryname"] == name:
                    log_file = config[log]["log_file"]
                    break
    if not name:
        for log in config:
            if "log_file" in config[log] and config[log]["log_file"] == log_file:
                if "entryname" in config[log]:
                    name = config[log]["entryname"]
                break

    # remove log if needed
    if log_file in config:
        res = __salt__["logadm.remove"](name if name else log_file)
        ret["result"] = "Error" not in res
        if ret["result"]:
            ret["comment"] = f"Configuration for {log_file} removed."
            ret["changes"][log_file] = None
        else:
            ret["comment"] = res["Error"]
    else:
        ret["result"] = True
        ret["comment"] = f"No configuration for {log_file} present."

    return ret



import getpass
import os
import shutil
import sys

import pytest

import salt.utils.files
import salt.utils.platform
import salt.utils.stringutils
from tests.support.case import ModuleCase
from tests.support.helpers import requires_system_grains
from tests.support.runtests import RUNTIME_VARS

# Posix only
try:
    import grp
    import pwd
except ImportError:
    pass

# Windows only
try:
    import win32file
except ImportError:
    pass


def symlink(source, link_name):
    """
    Handle symlinks on Windows with Python < 3.2
    """
    if salt.utils.platform.is_windows():
        win32file.CreateSymbolicLink(link_name, source)
    else:
        os.symlink(source, link_name)


@pytest.mark.windows_whitelisted
class FileModuleTest(ModuleCase):
    """
    Validate the file module
    """

    def setUp(self):
        self.myfile = os.path.join(RUNTIME_VARS.TMP, "myfile")
        with salt.utils.files.fopen(self.myfile, "w+") as fp:
            fp.write(salt.utils.stringutils.to_str("Hello" + os.linesep))
        self.mydir = os.path.join(RUNTIME_VARS.TMP, "mydir/isawesome")
        if not os.path.isdir(self.mydir):
            # left behind... Don't fail because of this!
            os.makedirs(self.mydir)
        self.mysymlink = os.path.join(RUNTIME_VARS.TMP, "mysymlink")
        if os.path.islink(self.mysymlink) or os.path.isfile(self.mysymlink):
            os.remove(self.mysymlink)
        symlink(self.myfile, self.mysymlink)
        self.mybadsymlink = os.path.join(RUNTIME_VARS.TMP, "mybadsymlink")
        if os.path.islink(self.mybadsymlink) or os.path.isfile(self.mybadsymlink):
            os.remove(self.mybadsymlink)
        symlink("/nonexistentpath", self.mybadsymlink)
        super().setUp()

    def tearDown(self):
        if os.path.isfile(self.myfile):
            os.remove(self.myfile)
        if os.path.islink(self.mysymlink) or os.path.isfile(self.mysymlink):
            os.remove(self.mysymlink)
        if os.path.islink(self.mybadsymlink) or os.path.isfile(self.mybadsymlink):
            os.remove(self.mybadsymlink)
        shutil.rmtree(self.mydir, ignore_errors=True)
        super().tearDown()

    @pytest.mark.skip_on_windows(reason="No security context on Windows")
    @requires_system_grains
    def test_get_selinux_context(self, grains):
        if grains.get("selinux", {}).get("enabled", False):
            NEW_CONTEXT = "system_u:object_r:system_conf_t:s0"
            self.run_function(
                "file.set_selinux_context", arg=[self.myfile, *(NEW_CONTEXT.split(":"))]
            )
            ret_file = self.run_function("file.get_selinux_context", arg=[self.myfile])
            self.assertEqual(ret_file, NEW_CONTEXT)

            # Issue #56557.  Ensure that the context of the directory
            # containing one file is the context of the directory itself, and
            # not the context of the first file in the directory.
            self.run_function(
                "file.set_selinux_context", arg=[self.mydir, *(NEW_CONTEXT.split(":"))]
            )
            ret_dir = self.run_function("file.get_selinux_context", arg=[self.mydir])
            self.assertEqual(ret_dir, NEW_CONTEXT)
            ret_updir = self.run_function(
                "file.get_selinux_context",
                arg=[os.path.abspath(os.path.join(self.mydir, ".."))],
            )
            self.assertNotEqual(ret_updir, NEW_CONTEXT)
        else:
            ret_file = self.run_function("file.get_selinux_context", arg=[self.myfile])
            self.assertIn("No selinux context information is available", ret_file)

    @pytest.mark.skip_on_windows(reason="No security context on Windows")
    @requires_system_grains
    def test_set_selinux_context(self, grains):
        if not grains.get("selinux", {}).get("enabled", False):
            self.skipTest("selinux not available")

        FILE_CONTEXT = "system_u:object_r:system_conf_t:s0"
        ret_file = self.run_function(
            "file.set_selinux_context", arg=[self.myfile, *(FILE_CONTEXT.split(":"))]
        )
        self.assertEqual(ret_file, FILE_CONTEXT)

        DIR_CONTEXT = "system_u:object_r:user_home_t:s0"
        ret_dir = self.run_function(
            "file.set_selinux_context", arg=[self.mydir, *(DIR_CONTEXT.split(":"))]
        )
        self.assertEqual(ret_dir, DIR_CONTEXT)

    @pytest.mark.skip_on_windows(reason="No chgrp on Windows")
    def test_chown(self):
        user = getpass.getuser()
        if sys.platform == "darwin":
            group = "staff"
        elif sys.platform.startswith(("linux", "freebsd", "openbsd")):
            group = grp.getgrgid(pwd.getpwuid(os.getuid()).pw_gid).gr_name
        ret = self.run_function("file.chown", arg=[self.myfile, user, group])
        self.assertIsNone(ret)
        fstat = os.stat(self.myfile)
        self.assertEqual(fstat.st_uid, os.getuid())
        self.assertEqual(fstat.st_gid, grp.getgrnam(group).gr_gid)

    @pytest.mark.skip_on_windows(reason="No chgrp on Windows")
    def test_chown_no_user(self):
        user = "notanyuseriknow"
        group = grp.getgrgid(pwd.getpwuid(os.getuid()).pw_gid).gr_name
        ret = self.run_function("file.chown", arg=[self.myfile, user, group])
        self.assertIn("not exist", ret)

    @pytest.mark.skip_on_windows(reason="No chgrp on Windows")
    def test_chown_no_user_no_group(self):
        user = "notanyuseriknow"
        group = "notanygroupyoushoulduse"
        ret = self.run_function("file.chown", arg=[self.myfile, user, group])
        self.assertIn("Group does not exist", ret)
        self.assertIn("User does not exist", ret)

    @pytest.mark.skip_on_windows(reason="No chgrp on Windows")
    def test_chown_no_path(self):
        user = getpass.getuser()
        if sys.platform == "darwin":
            group = "staff"
        elif sys.platform.startswith(("linux", "freebsd", "openbsd")):
            group = grp.getgrgid(pwd.getpwuid(os.getuid()).pw_gid).gr_name
        ret = self.run_function("file.chown", arg=["/tmp/nosuchfile", user, group])
        self.assertIn("File not found", ret)

    @pytest.mark.skip_on_windows(reason="No chgrp on Windows")
    def test_chown_noop(self):
        user = ""
        group = ""
        ret = self.run_function("file.chown", arg=[self.myfile, user, group])
        self.assertIsNone(ret)
        fstat = os.stat(self.myfile)
        self.assertEqual(fstat.st_uid, os.getuid())
        self.assertEqual(fstat.st_gid, os.getgid())

    @pytest.mark.skip_on_windows(reason="No chgrp on Windows")
    def test_chgrp(self):
        if sys.platform == "darwin":
            group = "everyone"
        elif sys.platform.startswith(("linux", "freebsd", "openbsd")):
            group = grp.getgrgid(pwd.getpwuid(os.getuid()).pw_gid).gr_name
        ret = self.run_function("file.chgrp", arg=[self.myfile, group])
        self.assertIsNone(ret)
        fstat = os.stat(self.myfile)
        self.assertEqual(fstat.st_gid, grp.getgrnam(group).gr_gid)

    @pytest.mark.skip_on_windows(reason="No chgrp on Windows")
    def test_chgrp_failure(self):
        group = "thisgroupdoesntexist"
        ret = self.run_function("file.chgrp", arg=[self.myfile, group])
        self.assertIn("not exist", ret)

    def test_patch(self):
        if not self.run_function("cmd.has_exec", ["patch"]):
            self.skipTest("patch is not installed")

        src_patch = os.path.join(RUNTIME_VARS.FILES, "file", "base", "hello.patch")
        src_file = os.path.join(RUNTIME_VARS.TMP, "src.txt")
        with salt.utils.files.fopen(src_file, "w+") as fp:
            fp.write(salt.utils.stringutils.to_str("Hello\n"))

        # dry-run should not modify src_file
        ret = self.minion_run("file.patch", src_file, src_patch, dry_run=True)
        assert ret["retcode"] == 0, repr(ret)
        with salt.utils.files.fopen(src_file) as fp:
            self.assertEqual(salt.utils.stringutils.to_unicode(fp.read()), "Hello\n")

        ret = self.minion_run("file.patch", src_file, src_patch)
        assert ret["retcode"] == 0, repr(ret)
        with salt.utils.files.fopen(src_file) as fp:
            self.assertEqual(
                salt.utils.stringutils.to_unicode(fp.read()), f"Hello world{os.linesep}"
            )

    def test_remove_file(self):
        ret = self.run_function("file.remove", arg=[self.myfile])
        self.assertTrue(ret)

    def test_remove_dir(self):
        ret = self.run_function("file.remove", arg=[self.mydir])
        self.assertTrue(ret)

    def test_remove_symlink(self):
        ret = self.run_function("file.remove", arg=[self.mysymlink])
        self.assertTrue(ret)

    def test_remove_broken_symlink(self):
        ret = self.run_function("file.remove", arg=[self.mybadsymlink])
        self.assertTrue(ret)

    def test_cannot_remove(self):
        ret = self.run_function("file.remove", arg=["tty"])
        self.assertEqual(
            "ERROR executing 'file.remove': File path must be absolute: tty", ret
        )

    def test_source_list_for_single_file_returns_unchanged(self):
        ret = self.run_function(
            "file.source_list", ["salt://http/httpd.conf", "filehash", "base"]
        )
        self.assertEqual(list(ret), ["salt://http/httpd.conf", "filehash"])

    def test_source_list_for_single_local_file_slash_returns_unchanged(self):
        ret = self.run_function("file.source_list", [self.myfile, "filehash", "base"])
        self.assertEqual(list(ret), [self.myfile, "filehash"])

    def test_source_list_for_single_local_file_proto_returns_unchanged(self):
        ret = self.run_function(
            "file.source_list", ["file://" + self.myfile, "filehash", "base"]
        )
        self.assertEqual(list(ret), ["file://" + self.myfile, "filehash"])

    def test_source_list_for_multiple_files_with_missing_files(self):
        file_list = [
            "salt://does/not/exist",
            "file://" + self.myfile,
            "http://localhost//does/not/exist",
            "salt://http/httpd.conf",
        ]
        ret = self.run_function("file.source_list", [file_list, "filehash", "base"])
        self.assertEqual(list(ret), ["file://" + self.myfile, "filehash"])

    def test_source_list_for_multiple_files_dict_with_missing_files(self):
        file_list = [
            {"salt://does/not/exist": "filehash"},
            {"file://" + self.myfile: "filehash"},
            {"http://localhost//does/not/exist": "filehash"},
            {"salt://http/httpd.conf": "filehash"},
        ]
        ret = self.run_function("file.source_list", [file_list, "", "base"])
        self.assertEqual(list(ret), ["file://" + self.myfile, "filehash"])

    def test_file_line_changes_format(self):
        """
        Test file.line changes output formatting.

        Issue #41474
        """
        ret = self.minion_run(
            "file.line", self.myfile, "Goodbye", mode="insert", after="Hello"
        )
        self.assertIn("Hello" + os.linesep + "+Goodbye", ret)

    def test_file_line_changes_entire_line(self):
        """
        Test file.line entire line matching

        Issue #49855
        """
        ret = self.minion_run(
            "file.line", self.myfile, "Goodbye", mode="insert", after="Hello"
        )
        assert "Hello" + os.linesep + "+Goodbye" in ret

        ret = self.minion_run(
            "file.line", self.myfile, "Goodbye 1", mode="insert", after="Hello"
        )
        assert (
            "Hello" + os.linesep + "+Goodbye 1" + os.linesep + " Goodbye" + os.linesep
            in ret
        )

        with salt.utils.files.fopen(self.myfile, "r") as fh_:
            content = fh_.read()

        assert (
            "Hello" + os.linesep + "Goodbye 1" + os.linesep + "Goodbye" + os.linesep
            == content
        )

    def test_file_line_content(self):
        self.minion_run(
            "file.line", self.myfile, "Goodbye", mode="insert", after="Hello"
        )
        with salt.utils.files.fopen(self.myfile, "r") as fp:
            content = fp.read()
        self.assertEqual(content, "Hello" + os.linesep + "Goodbye" + os.linesep)

    def test_file_line_duplicate_insert_after(self):
        """
        Test file.line duplicates line.

        Issue #50254
        """
        with salt.utils.files.fopen(self.myfile, "a") as fp:
            fp.write(salt.utils.stringutils.to_str("Goodbye" + os.linesep))
        self.minion_run(
            "file.line", self.myfile, "Goodbye", mode="insert", after="Hello"
        )
        with salt.utils.files.fopen(self.myfile, "r") as fp:
            content = fp.read()
        self.assertEqual(content, "Hello" + os.linesep + "Goodbye" + os.linesep)

    def test_file_line_duplicate_insert_before(self):
        """
        Test file.line duplicates line.

        Issue #50254
        """
        with salt.utils.files.fopen(self.myfile, "a") as fp:
            fp.write(salt.utils.stringutils.to_str("Goodbye" + os.linesep))
        self.minion_run(
            "file.line", self.myfile, "Hello", mode="insert", before="Goodbye"
        )
        with salt.utils.files.fopen(self.myfile, "r") as fp:
            content = fp.read()
        self.assertEqual(content, "Hello" + os.linesep + "Goodbye" + os.linesep)

    def test_file_line_duplicate_ensure_after(self):
        """
        Test file.line duplicates line.

        Issue #50254
        """
        with salt.utils.files.fopen(self.myfile, "a") as fp:
            fp.write(salt.utils.stringutils.to_str("Goodbye" + os.linesep))
        self.minion_run(
            "file.line", self.myfile, "Goodbye", mode="ensure", after="Hello"
        )
        with salt.utils.files.fopen(self.myfile, "r") as fp:
            content = fp.read()
        self.assertEqual(content, "Hello" + os.linesep + "Goodbye" + os.linesep)

    def test_file_line_duplicate_ensure_before(self):
        """
        Test file.line duplicates line.

        Issue #50254
        """
        with salt.utils.files.fopen(self.myfile, "a") as fp:
            fp.write(salt.utils.stringutils.to_str("Goodbye" + os.linesep))
        self.minion_run(
            "file.line", self.myfile, "Hello", mode="ensure", before="Goodbye"
        )
        with salt.utils.files.fopen(self.myfile, "r") as fp:
            content = fp.read()
        self.assertEqual(content, "Hello" + os.linesep + "Goodbye" + os.linesep)

    def test_file_read_bytes(self):
        """
        Test that ``file.read`` reads and returns ``bytes`` data
        """
        # Write some random bytes
        data = b"n\x1a\xf7S@tBI\xa9J"
        with salt.utils.files.fopen(self.myfile, "wb") as fp:
            fp.write(data)

        ret = self.minion_run("file.read", self.myfile, binary=True)
        self.assertEqual(type(ret), bytes)
        self.assertEqual(ret, data)

    def test_file_read_str(self):
        """
        Test that ``file.read`` reads and returns ``str`` data
        """
        # Write some str data
        data = "printable characters"
        with salt.utils.files.fopen(self.myfile, "w") as fp:
            fp.write(data)

        ret = self.minion_run("file.read", self.myfile)
        self.assertEqual(type(ret), str)
        self.assertEqual(ret, data)



"""
Manage Dell DRAC
"""

import logging

import salt.utils.path

log = logging.getLogger(__name__)


def __virtual__():
    if salt.utils.path.which("racadm"):
        return True

    return (
        False,
        "The drac execution module cannot be loaded: racadm binary not in path.",
    )


def __parse_drac(output):
    """
    Parse Dell DRAC output
    """
    drac = {}
    section = ""

    for i in output.splitlines():
        if i.rstrip() and "=" in i:
            if section in drac:
                drac[section].update(dict([[prop.strip() for prop in i.split("=")]]))
        else:
            section = i.strip()[:-1]
            if section not in drac and section:
                drac[section] = {}

    return drac


def __execute_cmd(command):
    """
    Execute rac commands
    """
    cmd = __salt__["cmd.run_all"](f"racadm {command}")

    if cmd["retcode"] != 0:
        log.warning("racadm return an exit code '%s'.", cmd["retcode"])
        return False

    return True


def system_info():
    """
    Return System information

    CLI Example:

    .. code-block:: bash

        salt dell drac.system_info
    """
    cmd = __salt__["cmd.run_all"]("racadm getsysinfo")

    if cmd["retcode"] != 0:
        log.warning("racadm return an exit code '%s'.", cmd["retcode"])

    return __parse_drac(cmd["stdout"])


def network_info():
    """
    Return Network Configuration

    CLI Example:

    .. code-block:: bash

        salt dell drac.network_info
    """

    cmd = __salt__["cmd.run_all"]("racadm getniccfg")

    if cmd["retcode"] != 0:
        log.warning("racadm return an exit code '%s'.", cmd["retcode"])

    return __parse_drac(cmd["stdout"])


def nameservers(*ns):
    """
    Configure the nameservers on the DRAC

    CLI Example:

    .. code-block:: bash

        salt dell drac.nameservers [NAMESERVERS]
        salt dell drac.nameservers ns1.example.com ns2.example.com
    """
    if len(ns) > 2:
        log.warning("racadm only supports two nameservers")
        return False

    for i in range(1, len(ns) + 1):
        if not __execute_cmd(
            f"config -g cfgLanNetworking -o cfgDNSServer{i} {ns[i - 1]}"
        ):
            return False

    return True


def syslog(server, enable=True):
    """
    Configure syslog remote logging, by default syslog will automatically be
    enabled if a server is specified. However, if you want to disable syslog
    you will need to specify a server followed by False

    CLI Example:

    .. code-block:: bash

        salt dell drac.syslog [SYSLOG IP] [ENABLE/DISABLE]
        salt dell drac.syslog 0.0.0.0 False
    """
    if enable and __execute_cmd("config -g cfgRemoteHosts -o cfgRhostsSyslogEnable 1"):
        return __execute_cmd(
            f"config -g cfgRemoteHosts -o cfgRhostsSyslogServer1 {server}"
        )

    return __execute_cmd("config -g cfgRemoteHosts -o cfgRhostsSyslogEnable 0")


def email_alerts(action):
    """
    Enable/Disable email alerts

    CLI Example:

    .. code-block:: bash

        salt dell drac.email_alerts True
        salt dell drac.email_alerts False
    """

    if action:
        return __execute_cmd("config -g cfgEmailAlert -o cfgEmailAlertEnable -i 1 1")
    else:
        return __execute_cmd("config -g cfgEmailAlert -o cfgEmailAlertEnable -i 1 0")


def list_users():
    """
    List all DRAC users

    CLI Example:

    .. code-block:: bash

        salt dell drac.list_users
    """
    users = {}
    _username = ""

    for idx in range(1, 17):
        cmd = __salt__["cmd.run_all"](f"racadm getconfig -g cfgUserAdmin -i {idx}")

        if cmd["retcode"] != 0:
            log.warning("racadm return an exit code '%s'.", cmd["retcode"])

        for user in cmd["stdout"].splitlines():
            if not user.startswith("cfg"):
                continue

            (key, val) = user.split("=")

            if key.startswith("cfgUserAdminUserName"):
                _username = val.strip()

                if val:
                    users[_username] = {"index": idx}
                else:
                    break
            else:
                users[_username].update({key: val})

    return users


def delete_user(username, uid=None):
    """
    Delete a user

    CLI Example:

    .. code-block:: bash

        salt dell drac.delete_user [USERNAME] [UID - optional]
        salt dell drac.delete_user diana 4
    """
    if uid is None:
        user = list_users()
        uid = user[username]["index"]

    if uid:
        return __execute_cmd(
            f'config -g cfgUserAdmin -o cfgUserAdminUserName -i {uid} ""'
        )

    else:
        log.warning("'%s' does not exist", username)
        return False

    return True


def change_password(username, password, uid=None):
    """
    Change users password

    CLI Example:

    .. code-block:: bash

        salt dell drac.change_password [USERNAME] [PASSWORD] [UID - optional]
        salt dell drac.change_password diana secret
    """
    if uid is None:
        user = list_users()
        uid = user[username]["index"]

    if uid:
        return __execute_cmd(
            "config -g cfgUserAdmin -o cfgUserAdminPassword -i {} {}".format(
                uid, password
            )
        )
    else:
        log.warning("'%s' does not exist", username)
        return False

    return True


def create_user(username, password, permissions, users=None):
    """
    Create user accounts

    CLI Example:

    .. code-block:: bash

        salt dell drac.create_user [USERNAME] [PASSWORD] [PRIVILEGES]
        salt dell drac.create_user diana secret login,test_alerts,clear_logs

    DRAC Privileges
      * login                   : Login to iDRAC
      * drac                    : Configure iDRAC
      * user_management         : Configure Users
      * clear_logs              : Clear Logs
      * server_control_commands : Execute Server Control Commands
      * console_redirection     : Access Console Redirection
      * virtual_media           : Access Virtual Media
      * test_alerts             : Test Alerts
      * debug_commands          : Execute Debug Commands
    """
    _uids = set()

    if users is None:
        users = list_users()

    if username in users:
        log.warning("'%s' already exists", username)
        return False

    for idx in users.keys():
        _uids.add(users[idx]["index"])

    uid = sorted(list(set(range(2, 12)) - _uids), reverse=True).pop()

    # Create user accountvfirst
    if not __execute_cmd(
        f"config -g cfgUserAdmin -o cfgUserAdminUserName -i {uid} {username}"
    ):
        delete_user(username, uid)
        return False

    # Configure users permissions
    if not set_permissions(username, permissions, uid):
        log.warning("unable to set user permissions")
        delete_user(username, uid)
        return False

    # Configure users password
    if not change_password(username, password, uid):
        log.warning("unable to set user password")
        delete_user(username, uid)
        return False

    # Enable users admin
    if not __execute_cmd(f"config -g cfgUserAdmin -o cfgUserAdminEnable -i {uid} 1"):
        delete_user(username, uid)
        return False

    return True


def set_permissions(username, permissions, uid=None):
    """
    Configure users permissions

    CLI Example:

    .. code-block:: bash

        salt dell drac.set_permissions [USERNAME] [PRIVILEGES] [USER INDEX - optional]
        salt dell drac.set_permissions diana login,test_alerts,clear_logs 4

    DRAC Privileges
      * login                   : Login to iDRAC
      * drac                    : Configure iDRAC
      * user_management         : Configure Users
      * clear_logs              : Clear Logs
      * server_control_commands : Execute Server Control Commands
      * console_redirection     : Access Console Redirection
      * virtual_media           : Access Virtual Media
      * test_alerts             : Test Alerts
      * debug_commands          : Execute Debug Commands
    """
    privileges = {
        "login": "0x0000001",
        "drac": "0x0000002",
        "user_management": "0x0000004",
        "clear_logs": "0x0000008",
        "server_control_commands": "0x0000010",
        "console_redirection": "0x0000020",
        "virtual_media": "0x0000040",
        "test_alerts": "0x0000080",
        "debug_commands": "0x0000100",
    }

    permission = 0

    # When users don't provide a user ID we need to search for this
    if uid is None:
        user = list_users()
        uid = user[username]["index"]

    # Generate privilege bit mask
    for i in permissions.split(","):
        perm = i.strip()

        if perm in privileges:
            permission += int(privileges[perm], 16)

    return __execute_cmd(
        "config -g cfgUserAdmin -o cfgUserAdminPrivilege -i {} 0x{:08X}".format(
            uid, permission
        )
    )


def set_snmp(community):
    """
    Configure SNMP community string

    CLI Example:

    .. code-block:: bash

        salt dell drac.set_snmp [COMMUNITY]
        salt dell drac.set_snmp public
    """
    return __execute_cmd(
        f"config -g cfgOobSnmp -o cfgOobSnmpAgentCommunity {community}"
    )


def set_network(ip, netmask, gateway):
    """
    Configure Network

    CLI Example:

    .. code-block:: bash

        salt dell drac.set_network [DRAC IP] [NETMASK] [GATEWAY]
        salt dell drac.set_network 192.168.0.2 255.255.255.0 192.168.0.1
    """
    return __execute_cmd(f"setniccfg -s {ip} {netmask} {gateway}")


def server_reboot():
    """
    Issues a power-cycle operation on the managed server. This action is
    similar to pressing the power button on the system's front panel to
    power down and then power up the system.

    CLI Example:

    .. code-block:: bash

        salt dell drac.server_reboot
    """
    return __execute_cmd("serveraction powercycle")


def server_poweroff():
    """
    Powers down the managed server.

    CLI Example:

    .. code-block:: bash

        salt dell drac.server_poweroff
    """
    return __execute_cmd("serveraction powerdown")


def server_poweron():
    """
    Powers up the managed server.

    CLI Example:

    .. code-block:: bash

        salt dell drac.server_poweron
    """
    return __execute_cmd("serveraction powerup")


def server_hardreset():
    """
    Performs a reset (reboot) operation on the managed server.

    CLI Example:

    .. code-block:: bash

        salt dell drac.server_hardreset
    """
    return __execute_cmd("serveraction hardreset")


def server_pxe():
    """
    Configure server to PXE perform a one off PXE boot

    CLI Example:

    .. code-block:: bash

        salt dell drac.server_pxe
    """
    if __execute_cmd("config -g cfgServerInfo -o cfgServerFirstBootDevice PXE"):
        if __execute_cmd("config -g cfgServerInfo -o cfgServerBootOnce 1"):
            return server_reboot
        else:
            log.warning("failed to set boot order")
            return False

    log.warning("failed to configure PXE boot")
    return False



import os
import pathlib
import shutil

import pytest

from tests.support.helpers import VirtualEnv
from tests.support.runtests import RUNTIME_VARS


@pytest.fixture(scope="package")
def setup_tests_path(tmp_path_factory):
    if os.environ.get("CI_RUN", "0") == "0":
        directory = tmp_path_factory.mktemp("setup-tests")
    else:
        # Under CI, on some platforms, Arch, Fedora 33, FreeBSD 12, we run out of disk space on /tmp
        # Use a subdirectory in the current user's home directory
        directory = pathlib.Path.home() / "setup-tests"
    directory.mkdir(parents=True, exist_ok=True)
    try:
        yield directory
    finally:
        shutil.rmtree(str(directory), ignore_errors=True)


@pytest.fixture
def virtualenv(setup_tests_path, pip_temp_dir):
    venv_dir = setup_tests_path / ".venv"
    try:
        yield VirtualEnv(venv_dir=venv_dir, env={"TMPDIR": str(pip_temp_dir)})
    finally:
        shutil.rmtree(str(venv_dir), ignore_errors=True)


@pytest.fixture
def cache_dir(setup_tests_path):
    _cache_dir = setup_tests_path / ".cache"
    _cache_dir.mkdir(parents=True, exist_ok=True)
    try:
        yield _cache_dir
    finally:
        shutil.rmtree(str(_cache_dir), ignore_errors=True)


@pytest.fixture(scope="package")
def pip_temp_dir(setup_tests_path):
    temp_dir_path = setup_tests_path / "temp-dir"
    temp_dir_path.mkdir()
    return temp_dir_path


@pytest.fixture(scope="package")
def src_dir(setup_tests_path):
    if os.environ.get("CI_RUN", "0") == "0":
        return RUNTIME_VARS.CODE_DIR

    _src_dir = setup_tests_path / "src"
    shutil.copytree(
        RUNTIME_VARS.CODE_DIR,
        str(_src_dir),
        ignore=shutil.ignore_patterns(
            "__pycache__", "*.pyc", "*.pyo", ".coverage.*", ".nox"
        ),
    )
    return str(_src_dir)



"""
The service module for FreeBSD

.. important::
    If you feel that Salt should be using this module to manage services on a
    minion, and it is using a different module (or gives an error similar to
    *'service.start' is not available*), see :ref:`here
    <module-provider-override>`.
"""

import fnmatch
import logging
import os
import re

import salt.utils.decorators as decorators
import salt.utils.files
import salt.utils.path
from salt.exceptions import CommandNotFoundError

__func_alias__ = {"reload_": "reload"}

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "service"


def __virtual__():
    """
    Only work on FreeBSD
    """
    # Disable on these platforms, specific service modules exist:
    if __grains__["os"] == "FreeBSD":
        return __virtualname__
    return (
        False,
        "The freebsdservice execution module cannot be loaded: only available on"
        " FreeBSD systems.",
    )


@decorators.memoize
def _cmd(jail=None):
    """
    Return full path to service command

    .. versionchanged:: 2016.3.4

    Support for jail (representing jid or jail name) keyword argument in kwargs
    """
    service = salt.utils.path.which("service")
    if not service:
        raise CommandNotFoundError("'service' command not found")
    if jail:
        jexec = salt.utils.path.which("jexec")
        if not jexec:
            raise CommandNotFoundError("'jexec' command not found")
        service = f"{jexec} {jail} {service}"
    return service


def _get_jail_path(jail):
    """
    .. versionadded:: 2016.3.4

    Return the jail's root directory (path) as shown in jls

    jail
        The jid or jail name
    """
    jls = salt.utils.path.which("jls")
    if not jls:
        raise CommandNotFoundError("'jls' command not found")
    jails = __salt__["cmd.run_stdout"](f"{jls} -n jid name path")
    for j in jails.splitlines():
        jid, jname, path = (x.split("=")[1].strip() for x in j.split())
        if jid == jail or jname == jail:
            return path.rstrip("/")
    # XΧΧ, TODO, not sure how to handle nonexistent jail
    return ""


def _get_rcscript(name, jail=None):
    """
    Return full path to service rc script

    .. versionchanged:: 2016.3.4

    Support for jail (representing jid or jail name) keyword argument in kwargs
    """
    cmd = f"{_cmd(jail)} -r"
    prf = _get_jail_path(jail) if jail else ""
    for line in __salt__["cmd.run_stdout"](cmd, python_shell=False).splitlines():
        if line.endswith(f"{os.path.sep}{name}"):
            return os.path.join(prf, line.lstrip(os.path.sep))
    return None


def _get_rcvar(name, jail=None):
    """
    Return rcvar

    .. versionchanged:: 2016.3.4

    Support for jail (representing jid or jail name) keyword argument in kwargs
    """
    if not available(name, jail):
        log.error("Service %s not found", name)
        return False

    cmd = f"{_cmd(jail)} {name} rcvar"

    for line in __salt__["cmd.run_stdout"](cmd, python_shell=False).splitlines():
        if '_enable="' not in line:
            continue
        rcvar, _ = line.split("=", 1)
        return rcvar

    return None


def get_enabled(jail=None):
    """
    Return what services are set to run on boot

    .. versionchanged:: 2016.3.4

    Support for jail (representing jid or jail name) keyword argument in kwargs

    CLI Example:

    .. code-block:: bash

        salt '*' service.get_enabled
    """
    ret = []
    service = _cmd(jail)
    prf = _get_jail_path(jail) if jail else ""
    for svc in __salt__["cmd.run"](f"{service} -e").splitlines():
        ret.append(os.path.basename(svc))

    # This is workaround for bin/173454 bug
    for svc in get_all(jail):
        if svc in ret:
            continue
        if not os.path.exists(f"{prf}/etc/rc.conf.d/{svc}"):
            continue
        if enabled(svc, jail=jail):
            ret.append(svc)

    return sorted(ret)


def get_disabled(jail=None):
    """
    Return what services are available but not enabled to start at boot

    .. versionchanged:: 2016.3.4

    Support for jail (representing jid or jail name) keyword argument in kwargs

    CLI Example:

    .. code-block:: bash

        salt '*' service.get_disabled
    """
    en_ = get_enabled(jail)
    all_ = get_all(jail)
    return sorted(set(all_) - set(en_))


def _switch(name, on, **kwargs):  # pylint: disable=C0103  # pylint: disable=C0103
    """
    Switch on/off service start at boot.

    .. versionchanged:: 2016.3.4

    Support for jail (representing jid or jail name) and chroot keyword argument
    in kwargs. chroot should be used when jail's /etc is mounted read-only and
    should point to a root directory where jail's /etc is mounted read-write.
    """
    jail = kwargs.get("jail", "")
    chroot = kwargs.get("chroot", "").rstrip("/")
    if not available(name, jail):
        return False

    rcvar = _get_rcvar(name, jail)
    if not rcvar:
        log.error("rcvar for service %s not found", name)
        return False

    if jail and not chroot:
        # prepend the jail's path in config paths when referring to a jail, when
        # chroot is not provided. chroot should be provided when the jail's /etc
        # is mounted read-only
        chroot = _get_jail_path(jail)

    config = kwargs.get(
        "config",
        __salt__["config.option"]("service.config", default=f"{chroot}/etc/rc.conf"),
    )

    if not config:
        rcdir = f"{chroot}/etc/rc.conf.d"
        if not os.path.exists(rcdir) or not os.path.isdir(rcdir):
            log.error("%s not exists", rcdir)
            return False
        config = os.path.join(rcdir, rcvar.replace("_enable", ""))

    nlines = []
    edited = False

    if on:
        val = "YES"
    else:
        val = "NO"

    if os.path.exists(config):
        with salt.utils.files.fopen(config, "r") as ifile:
            for line in ifile:
                line = salt.utils.stringutils.to_unicode(line)
                if not line.startswith(f"{rcvar}="):
                    nlines.append(line)
                    continue
                rest = line[len(line.split()[0]) :]  # keep comments etc
                nlines.append(f'{rcvar}="{val}"{rest}')
                edited = True
    if not edited:
        # Ensure that the file ends in a \n
        if len(nlines) > 1 and nlines[-1][-1] != "\n":
            nlines[-1] = f"{nlines[-1]}\n"
        nlines.append(f'{rcvar}="{val}"\n')

    with salt.utils.files.fopen(config, "w") as ofile:
        nlines = [salt.utils.stringutils.to_str(_l) for _l in nlines]
        ofile.writelines(nlines)

    return True


def enable(name, **kwargs):
    """
    Enable the named service to start at boot

    name
        service name

    config : /etc/rc.conf
        Config file for managing service. If config value is
        empty string, then /etc/rc.conf.d/<service> used.
        See man rc.conf(5) for details.

        Also service.config variable can be used to change default.

    .. versionchanged:: 2016.3.4

    jail (optional keyword argument)
        the jail's id or name

    chroot (optional keyword argument)
        the jail's chroot, if the jail's /etc is not mounted read-write

    CLI Example:

    .. code-block:: bash

        salt '*' service.enable <service name>
    """
    return _switch(name, True, **kwargs)


def disable(name, **kwargs):
    """
    Disable the named service to start at boot

    Arguments the same as for enable()

    .. versionchanged:: 2016.3.4

    jail (optional keyword argument)
        the jail's id or name

    chroot (optional keyword argument)
        the jail's chroot, if the jail's /etc is not mounted read-write

    CLI Example:

    .. code-block:: bash

        salt '*' service.disable <service name>
    """
    return _switch(name, False, **kwargs)


def enabled(name, **kwargs):
    """
    Return True if the named service is enabled, false otherwise

    name
        Service name

    .. versionchanged:: 2016.3.4

    Support for jail (representing jid or jail name) keyword argument in kwargs

    CLI Example:

    .. code-block:: bash

        salt '*' service.enabled <service name>
    """
    jail = kwargs.get("jail", "")
    if not available(name, jail):
        log.error("Service %s not found", name)
        return False

    cmd = f"{_cmd(jail)} {name} rcvar"

    for line in __salt__["cmd.run_stdout"](cmd, python_shell=False).splitlines():
        if '_enable="' not in line:
            continue
        _, state, _ = line.split('"', 2)
        return state.lower() in ("yes", "true", "on", "1")

    # probably will never reached
    return False


def disabled(name, **kwargs):
    """
    Return True if the named service is enabled, false otherwise

    CLI Example:

    .. code-block:: bash

        salt '*' service.disabled <service name>
    """
    return not enabled(name, **kwargs)


def available(name, jail=None):
    """
    Check that the given service is available.

    .. versionchanged:: 2016.3.4

    jail: optional jid or jail name

    CLI Example:

    .. code-block:: bash

        salt '*' service.available sshd
    """
    return name in get_all(jail)


def missing(name, jail=None):
    """
    The inverse of service.available.
    Returns ``True`` if the specified service is not available, otherwise returns
    ``False``.

    .. versionchanged:: 2016.3.4

    jail: optional jid or jail name

    CLI Example:

    .. code-block:: bash

        salt '*' service.missing sshd
    """
    return name not in get_all(jail)


def get_all(jail=None):
    """
    Return a list of all available services

    .. versionchanged:: 2016.3.4

    jail: optional jid or jail name

    CLI Example:

    .. code-block:: bash

        salt '*' service.get_all
    """
    ret = []
    service = _cmd(jail)
    for srv in __salt__["cmd.run"](f"{service} -l").splitlines():
        if not srv.isupper():
            ret.append(srv)
    return sorted(ret)


def start(name, jail=None):
    """
    Start the specified service

    .. versionchanged:: 2016.3.4

    jail: optional jid or jail name

    CLI Example:

    .. code-block:: bash

        salt '*' service.start <service name>
    """
    cmd = f"{_cmd(jail)} {name} onestart"
    return not __salt__["cmd.retcode"](cmd, python_shell=False)


def stop(name, jail=None):
    """
    Stop the specified service

    .. versionchanged:: 2016.3.4

    jail: optional jid or jail name

    CLI Example:

    .. code-block:: bash

        salt '*' service.stop <service name>
    """
    cmd = f"{_cmd(jail)} {name} onestop"
    return not __salt__["cmd.retcode"](cmd, python_shell=False)


def restart(name, jail=None):
    """
    Restart the named service

    .. versionchanged:: 2016.3.4

    jail: optional jid or jail name

    CLI Example:

    .. code-block:: bash

        salt '*' service.restart <service name>
    """
    cmd = f"{_cmd(jail)} {name} onerestart"
    return not __salt__["cmd.retcode"](cmd, python_shell=False)


def reload_(name, jail=None):
    """
    Restart the named service

    .. versionchanged:: 2016.3.4

    jail: optional jid or jail name

    CLI Example:

    .. code-block:: bash

        salt '*' service.reload <service name>
    """
    cmd = f"{_cmd(jail)} {name} onereload"
    return not __salt__["cmd.retcode"](cmd, python_shell=False)


def status(name, sig=None, jail=None):
    """
    Return the status for a service.
    If the name contains globbing, a dict mapping service name to True/False
    values is returned.

    .. versionchanged:: 2016.3.4

    .. versionchanged:: 2018.3.0
        The service name can now be a glob (e.g. ``salt*``)

    Args:
        name (str): The name of the service to check
        sig (str): Signature to use to find the service via ps

    Returns:
        bool: True if running, False otherwise
        dict: Maps service name to True if running, False otherwise

    CLI Example:

    .. code-block:: bash

        salt '*' service.status <service name> [service signature]
    """
    if sig:
        return bool(__salt__["status.pid"](sig))

    contains_globbing = bool(re.search(r"\*|\?|\[.+\]", name))
    if contains_globbing:
        services = fnmatch.filter(get_all(), name)
    else:
        services = [name]
    results = {}
    for service in services:
        cmd = f"{_cmd(jail)} {service} onestatus"
        results[service] = not __salt__["cmd.retcode"](
            cmd, python_shell=False, ignore_retcode=True
        )
    if contains_globbing:
        return results
    return results[name]



"""
Utilities to enable exception reraising across the master commands

"""

import builtins

import salt.exceptions
import salt.utils.event


def raise_error(name=None, args=None, message=""):
    """
    Raise an exception with __name__ from name, args from args
    If args is None Otherwise message from message\
    If name is empty then use "Exception"
    """
    name = name or "Exception"
    if hasattr(salt.exceptions, name):
        ex = getattr(salt.exceptions, name)
    elif hasattr(builtins, name):
        ex = getattr(builtins, name)
    else:
        name = "SaltException"
        ex = getattr(salt.exceptions, name)
    if args is not None:
        raise ex(*args)
    else:
        raise ex(message)


def pack_exception(exc):
    if hasattr(exc, "pack"):
        packed_exception = exc.pack()
    else:
        packed_exception = {"message": exc.__unicode__(), "args": exc.args}
    return packed_exception


def fire_exception(exc, opts, job=None, node="minion"):
    """
    Fire raw exception across the event bus
    """
    if job is None:
        job = {}
    event = salt.utils.event.SaltEvent(node, opts=opts, listen=False)
    event.fire_event(pack_exception(exc), "_salt_error")



"""
Utility functions for SMB connections

:depends: impacket
"""

import logging
import socket
import uuid

import salt.utils.files
import salt.utils.stringutils
import salt.utils.versions
from salt.exceptions import MissingSmb

log = logging.getLogger(__name__)


try:
    from smbprotocol.connection import Connection
    from smbprotocol.create_contexts import (
        CreateContextName,
        SMB2CreateContextRequest,
        SMB2CreateQueryMaximalAccessRequest,
    )
    from smbprotocol.open import (
        CreateDisposition,
        CreateOptions,
        DirectoryAccessMask,
        FileAttributes,
        FileInformationClass,
        FilePipePrinterAccessMask,
        ImpersonationLevel,
        Open,
        ShareAccess,
    )
    from smbprotocol.security_descriptor import (
        AccessAllowedAce,
        AccessMask,
        AclPacket,
        SDControl,
        SIDPacket,
        SMB2CreateSDBuffer,
    )
    from smbprotocol.session import Session
    from smbprotocol.tree import TreeConnect

    logging.getLogger("smbprotocol").setLevel(logging.WARNING)
    HAS_SMBPROTOCOL = True
except ImportError:
    HAS_SMBPROTOCOL = False


class SMBProto:
    def __init__(self, server, username, password, port=445):
        connection_id = uuid.uuid4()
        addr = socket.getaddrinfo(server, None, 0, 0, socket.IPPROTO_TCP)[0][4][0]
        self.server = server
        connection = Connection(connection_id, addr, port, require_signing=True)
        self.session = Session(connection, username, password, require_encryption=False)

    def connect(self):
        self.connection.connect()
        self.session.connect()

    def close(self):
        self.session.connection.disconnect(True)

    @property
    def connection(self):
        return self.session.connection

    def tree_connect(self, share):
        if share.endswith("$"):
            share = rf"\\{self.server}\{share}"
        tree = TreeConnect(self.session, share)
        tree.connect()
        return tree

    @staticmethod
    def normalize_filename(file):
        return file.lstrip("\\")

    @classmethod
    def open_file(cls, tree, file):
        file = cls.normalize_filename(file)
        # ensure file is created, get maximal access, and set everybody read access
        max_req = SMB2CreateContextRequest()
        max_req["buffer_name"] = (
            CreateContextName.SMB2_CREATE_QUERY_MAXIMAL_ACCESS_REQUEST
        )
        max_req["buffer_data"] = SMB2CreateQueryMaximalAccessRequest()

        # create security buffer that sets the ACL for everyone to have read access
        everyone_sid = SIDPacket()
        everyone_sid.from_string("S-1-1-0")
        ace = AccessAllowedAce()
        ace["mask"] = AccessMask.GENERIC_ALL
        ace["sid"] = everyone_sid
        acl = AclPacket()
        acl["aces"] = [ace]
        sec_desc = SMB2CreateSDBuffer()
        sec_desc["control"].set_flag(SDControl.SELF_RELATIVE)
        sec_desc.set_dacl(acl)
        sd_buffer = SMB2CreateContextRequest()
        sd_buffer["buffer_name"] = CreateContextName.SMB2_CREATE_SD_BUFFER
        sd_buffer["buffer_data"] = sec_desc

        create_contexts = [max_req, sd_buffer]
        file_open = Open(tree, file)
        open_info = file_open.create(
            ImpersonationLevel.Impersonation,
            FilePipePrinterAccessMask.GENERIC_READ
            | FilePipePrinterAccessMask.GENERIC_WRITE,
            FileAttributes.FILE_ATTRIBUTE_NORMAL,
            ShareAccess.FILE_SHARE_READ | ShareAccess.FILE_SHARE_WRITE,
            CreateDisposition.FILE_OVERWRITE_IF,
            CreateOptions.FILE_NON_DIRECTORY_FILE,
        )
        return file_open

    @staticmethod
    def open_directory(tree, name, create=False):
        # ensure directory is created
        dir_open = Open(tree, name)
        if create:
            dir_open.create(
                ImpersonationLevel.Impersonation,
                DirectoryAccessMask.GENERIC_READ | DirectoryAccessMask.GENERIC_WRITE,
                FileAttributes.FILE_ATTRIBUTE_DIRECTORY,
                ShareAccess.FILE_SHARE_READ | ShareAccess.FILE_SHARE_WRITE,
                CreateDisposition.FILE_OPEN_IF,
                CreateOptions.FILE_DIRECTORY_FILE,
            )
        return dir_open


def _get_conn_smbprotocol(host="", username="", password="", client_name="", port=445):
    conn = SMBProto(host, username, password, port)
    conn.connect()
    return conn


def get_conn(host="", username=None, password=None, port=445):
    """
    Get an SMB connection
    """
    if HAS_SMBPROTOCOL:
        log.info("Get connection smbprotocol")
        return _get_conn_smbprotocol(host, username, password, port=port)
    else:
        return False


def _mkdirs_smbprotocol(
    path, share="C$", conn=None, host=None, username=None, password=None
):
    if conn is None:
        conn = get_conn(host, username, password)

    if conn is False:
        return False

    tree = conn.tree_connect(share)
    comps = path.split("/")
    pos = 1
    for comp in comps:
        cwd = "\\".join(comps[0:pos])
        dir_open = conn.open_directory(tree, cwd, create=True)
        compound_messages = [
            dir_open.query_directory(
                "*", FileInformationClass.FILE_NAMES_INFORMATION, send=False
            ),
            dir_open.close(False, send=False),
        ]
        requests = conn.session.connection.send_compound(
            [x[0] for x in compound_messages],
            conn.session.session_id,
            tree.tree_connect_id,
        )
        for i, request in enumerate(requests):
            response = compound_messages[i][1](request)
        pos += 1


def mkdirs(path, share="C$", conn=None, host=None, username=None, password=None):
    if HAS_SMBPROTOCOL:
        return _mkdirs_smbprotocol(
            path, share, conn=conn, host=host, username=username, password=password
        )
    raise MissingSmb("SMB library required (impacket or smbprotocol)")


def _put_str_smbprotocol(
    content, path, share="C$", conn=None, host=None, username=None, password=None
):
    if conn is None:
        conn = get_conn(host, username, password)
    if conn is False:
        return False
    tree = conn.tree_connect(share)
    try:
        file_open = conn.open_file(tree, path)
        file_open.write(salt.utils.stringutils.to_bytes(content), 0)
    finally:
        file_open.close()


def put_str(
    content, path, share="C$", conn=None, host=None, username=None, password=None
):
    """
    Wrapper around impacket.smbconnection.putFile() that allows a string to be
    uploaded, without first writing it as a local file
    """
    if HAS_SMBPROTOCOL:
        return _put_str_smbprotocol(
            content,
            path,
            share,
            conn=conn,
            host=host,
            username=username,
            password=password,
        )
    raise MissingSmb("SMB library required (impacket or smbprotocol)")


def _put_file_smbprotocol(
    local_path,
    path,
    share="C$",
    conn=None,
    host=None,
    username=None,
    password=None,
    chunk_size=1024 * 1024,
):
    if conn is None:
        conn = get_conn(host, username, password)
    if conn is False:
        return False

    tree = conn.tree_connect(share)
    file_open = conn.open_file(tree, path)
    with salt.utils.files.fopen(local_path, "rb") as fh_:
        try:
            position = 0
            while True:
                chunk = fh_.read(chunk_size)
                if not chunk:
                    break
                file_open.write(chunk, position)
                position += len(chunk)
        finally:
            file_open.close(False)


def put_file(
    local_path, path, share="C$", conn=None, host=None, username=None, password=None
):
    """
    Wrapper around impacket.smbconnection.putFile() that allows a file to be
    uploaded

    Example usage:

        import salt.utils.smb
        smb_conn = salt.utils.smb.get_conn('10.0.0.45', 'vagrant', 'vagrant')
        salt.utils.smb.put_file('/root/test.pdf', 'temp\\myfiles\\test1.pdf', conn=smb_conn)
    """
    if HAS_SMBPROTOCOL:
        return _put_file_smbprotocol(
            local_path,
            path,
            share,
            conn=conn,
            host=host,
            username=username,
            password=password,
        )
    raise MissingSmb("SMB library required (impacket or smbprotocol)")


def _delete_file_smbprotocol(
    path, share="C$", conn=None, host=None, username=None, password=None
):
    if conn is None:
        conn = get_conn(host, username, password)
    if conn is False:
        return False
    tree = conn.tree_connect(share)
    file_open = Open(tree, path)
    delete_msgs = [
        file_open.create(
            ImpersonationLevel.Impersonation,
            FilePipePrinterAccessMask.GENERIC_READ | FilePipePrinterAccessMask.DELETE,
            FileAttributes.FILE_ATTRIBUTE_NORMAL,
            ShareAccess.FILE_SHARE_READ | ShareAccess.FILE_SHARE_WRITE,
            CreateDisposition.FILE_OPEN,
            CreateOptions.FILE_NON_DIRECTORY_FILE | CreateOptions.FILE_DELETE_ON_CLOSE,
            send=False,
        ),
        file_open.close(False, send=False),
    ]
    requests = conn.connection.send_compound(
        [x[0] for x in delete_msgs],
        conn.session.session_id,
        tree.tree_connect_id,
        related=True,
    )
    responses = []
    for i, request in enumerate(requests):
        # A SMBResponseException will be raised if something went wrong
        response = delete_msgs[i][1](request)
        responses.append(response)


def delete_file(path, share="C$", conn=None, host=None, username=None, password=None):
    if HAS_SMBPROTOCOL:
        return _delete_file_smbprotocol(
            path, share, conn=conn, host=host, username=username, password=password
        )
    raise MissingSmb("SMB library required (impacket or smbprotocol)")


def _delete_directory_smbprotocol(
    path, share="C$", conn=None, host=None, username=None, password=None
):
    if conn is None:
        conn = get_conn(host, username, password)
    if conn is False:
        return False
    log.debug("_delete_directory_smbprotocol - share: %s, path: %s", share, path)
    tree = conn.tree_connect(share)

    dir_open = Open(tree, path)
    delete_msgs = [
        dir_open.create(
            ImpersonationLevel.Impersonation,
            DirectoryAccessMask.DELETE,
            FileAttributes.FILE_ATTRIBUTE_DIRECTORY,
            0,
            CreateDisposition.FILE_OPEN,
            CreateOptions.FILE_DIRECTORY_FILE | CreateOptions.FILE_DELETE_ON_CLOSE,
            send=False,
        ),
        dir_open.close(False, send=False),
    ]
    delete_reqs = conn.connection.send_compound(
        [x[0] for x in delete_msgs],
        sid=conn.session.session_id,
        tid=tree.tree_connect_id,
        related=True,
    )
    for i, request in enumerate(delete_reqs):
        # A SMBResponseException will be raised if something went wrong
        response = delete_msgs[i][1](request)


def delete_directory(
    path, share="C$", conn=None, host=None, username=None, password=None
):
    if HAS_SMBPROTOCOL:
        return _delete_directory_smbprotocol(
            path, share, conn=conn, host=host, username=username, password=password
        )
    raise MissingSmb("SMB library required (impacket or smbprotocol)")



"""
Connection module for Amazon S3 Buckets

.. versionadded:: 2016.3.0

:depends:
    - boto
    - boto3

The dependencies listed above can be installed via package or pip.

:configuration: This module accepts explicit Lambda credentials but can also
    utilize IAM roles assigned to the instance through Instance Profiles.
    Dynamic credentials are then automatically obtained from AWS API and no
    further configuration is necessary. More Information available at:

    .. code-block:: text

        http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/iam-roles-for-amazon-ec2.html

    If IAM roles are not used you need to specify them either in a pillar or
    in the minion's config file:

    .. code-block:: yaml

        s3.keyid: GKTADJGHEIQSXMKKRBJ08H
        s3.key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

    A region may also be specified in the configuration:

    .. code-block:: yaml

        s3.region: us-east-1

    If a region is not specified, the default is us-east-1.

    It's also possible to specify key, keyid and region via a profile, either
    as a passed in dict, or as a string to pull from pillars or minion config:

    .. code-block:: yaml

        myprofile:
            keyid: GKTADJGHEIQSXMKKRBJ08H
            key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
            region: us-east-1

"""

# keep lint from choking on _get_conn and _cache_id
# pylint: disable=E0602
#  disable complaints about perfectly valid non-assignment code
# pylint: disable=W0106


import logging

import salt.utils.compat
import salt.utils.json
import salt.utils.versions
from salt.exceptions import SaltInvocationError

log = logging.getLogger(__name__)


# pylint: disable=import-error
try:
    # pylint: disable=unused-import
    import boto
    import boto3

    # pylint: enable=unused-import
    from botocore.exceptions import ClientError

    logging.getLogger("boto3").setLevel(logging.CRITICAL)
    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False
# pylint: enable=import-error


def __virtual__():
    """
    Only load if boto libraries exist and if boto libraries are greater than
    a given version.
    """
    # the boto_lambda execution module relies on the connect_to_region() method
    # which was added in boto 2.8.0
    # https://github.com/boto/boto/commit/33ac26b416fbb48a60602542b4ce15dcc7029f12
    return salt.utils.versions.check_boto_reqs(boto3_ver="1.2.1")


def __init__(opts):
    if HAS_BOTO:
        __utils__["boto3.assign_funcs"](__name__, "s3")


def exists(Bucket, region=None, key=None, keyid=None, profile=None):
    """
    Given a bucket name, check to see if the given bucket exists.

    Returns True if the given bucket exists and returns False if the given
    bucket does not exist.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_s3_bucket.exists mybucket

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        buckets = conn.head_bucket(Bucket=Bucket)
        return {"exists": True}
    except ClientError as e:
        if e.response.get("Error", {}).get("Code") == "404":
            return {"exists": False}
        err = __utils__["boto3.get_error"](e)
        return {"error": err}


def create(
    Bucket,
    ACL=None,
    LocationConstraint=None,
    GrantFullControl=None,
    GrantRead=None,
    GrantReadACP=None,
    GrantWrite=None,
    GrantWriteACP=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Given a valid config, create an S3 Bucket.

    Returns {created: true} if the bucket was created and returns
    {created: False} if the bucket was not created.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_s3_bucket.create my_bucket \\
                         GrantFullControl='emailaddress=example@example.com' \\
                         GrantRead='uri="http://acs.amazonaws.com/groups/global/AllUsers"' \\
                         GrantReadACP='emailaddress="exampl@example.com",id="2345678909876432"' \\
                         LocationConstraint=us-west-1

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        kwargs = {}
        for arg in (
            "ACL",
            "GrantFullControl",
            "GrantRead",
            "GrantReadACP",
            "GrantWrite",
            "GrantWriteACP",
        ):
            if locals()[arg] is not None:
                kwargs[arg] = str(locals()[arg])
        if LocationConstraint:
            kwargs["CreateBucketConfiguration"] = {
                "LocationConstraint": LocationConstraint
            }
        location = conn.create_bucket(Bucket=Bucket, **kwargs)
        conn.get_waiter("bucket_exists").wait(Bucket=Bucket)
        if location:
            log.info(
                "The newly created bucket name is located at %s", location["Location"]
            )

            return {"created": True, "name": Bucket, "Location": location["Location"]}
        else:
            log.warning("Bucket was not created")
            return {"created": False}
    except ClientError as e:
        return {"created": False, "error": __utils__["boto3.get_error"](e)}


def delete(
    Bucket,
    MFA=None,
    RequestPayer=None,
    Force=False,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Given a bucket name, delete it, optionally emptying it first.

    Returns {deleted: true} if the bucket was deleted and returns
    {deleted: false} if the bucket was not deleted.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_s3_bucket.delete mybucket

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        if Force:
            empty(
                Bucket,
                MFA=MFA,
                RequestPayer=RequestPayer,
                region=region,
                key=key,
                keyid=keyid,
                profile=profile,
            )
        conn.delete_bucket(Bucket=Bucket)
        return {"deleted": True}
    except ClientError as e:
        return {"deleted": False, "error": __utils__["boto3.get_error"](e)}


def delete_objects(
    Bucket,
    Delete,
    MFA=None,
    RequestPayer=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Delete objects in a given S3 bucket.

    Returns {deleted: true} if all objects were deleted
    and {deleted: false, failed: [key, ...]} otherwise

    CLI Example:

    .. code-block:: bash

        salt myminion boto_s3_bucket.delete_objects mybucket '{Objects: [Key: myobject]}'

    """

    if isinstance(Delete, str):
        Delete = salt.utils.json.loads(Delete)
    if not isinstance(Delete, dict):
        raise SaltInvocationError("Malformed Delete request.")
    if "Objects" not in Delete:
        raise SaltInvocationError("Malformed Delete request.")

    failed = []
    objs = Delete["Objects"]
    for i in range(0, len(objs), 1000):
        chunk = objs[i : i + 1000]
        subset = {"Objects": chunk, "Quiet": True}
        try:
            args = {"Bucket": Bucket}
            args.update({"MFA": MFA}) if MFA else None
            args.update({"RequestPayer": RequestPayer}) if RequestPayer else None
            args.update({"Delete": subset})
            conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
            ret = conn.delete_objects(**args)
            failed += ret.get("Errors", [])
        except ClientError as e:
            return {"deleted": False, "error": __utils__["boto3.get_error"](e)}

    if failed:
        return {"deleted": False, "failed": failed}
    else:
        return {"deleted": True}


def describe(Bucket, region=None, key=None, keyid=None, profile=None):
    """
    Given a bucket name describe its properties.

    Returns a dictionary of interesting properties.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_s3_bucket.describe mybucket

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        result = {}
        conn_dict = {
            "ACL": conn.get_bucket_acl,
            "CORS": conn.get_bucket_cors,
            "LifecycleConfiguration": conn.get_bucket_lifecycle_configuration,
            "Location": conn.get_bucket_location,
            "Logging": conn.get_bucket_logging,
            "NotificationConfiguration": conn.get_bucket_notification_configuration,
            "Policy": conn.get_bucket_policy,
            "Replication": conn.get_bucket_replication,
            "RequestPayment": conn.get_bucket_request_payment,
            "Versioning": conn.get_bucket_versioning,
            "Website": conn.get_bucket_website,
        }

        for key, query in conn_dict.items():
            try:
                data = query(Bucket=Bucket)
            except ClientError as e:
                if e.response.get("Error", {}).get("Code") in (
                    "NoSuchLifecycleConfiguration",
                    "NoSuchCORSConfiguration",
                    "NoSuchBucketPolicy",
                    "NoSuchWebsiteConfiguration",
                    "ReplicationConfigurationNotFoundError",
                    "NoSuchTagSet",
                ):
                    continue
                raise
            if "ResponseMetadata" in data:
                del data["ResponseMetadata"]
            result[key] = data

        tags = {}
        try:
            data = conn.get_bucket_tagging(Bucket=Bucket)
            for tagdef in data.get("TagSet"):
                tags[tagdef.get("Key")] = tagdef.get("Value")
        except ClientError as e:
            if not e.response.get("Error", {}).get("Code") == "NoSuchTagSet":
                raise
        if tags:
            result["Tagging"] = tags
        return {"bucket": result}
    except ClientError as e:
        err = __utils__["boto3.get_error"](e)
        if e.response.get("Error", {}).get("Code") == "NoSuchBucket":
            return {"bucket": None}
        return {"error": __utils__["boto3.get_error"](e)}


def empty(
    Bucket, MFA=None, RequestPayer=None, region=None, key=None, keyid=None, profile=None
):
    """
    Delete all objects in a given S3 bucket.

    Returns {deleted: true} if all objects were deleted
    and {deleted: false, failed: [key, ...]} otherwise

    CLI Example:

    .. code-block:: bash

        salt myminion boto_s3_bucket.empty mybucket

    """

    stuff = list_object_versions(
        Bucket, region=region, key=key, keyid=keyid, profile=profile
    )
    Delete = {}
    Delete["Objects"] = [
        {"Key": v["Key"], "VersionId": v["VersionId"]}
        for v in stuff.get("Versions", [])
    ]
    Delete["Objects"] += [
        {"Key": v["Key"], "VersionId": v["VersionId"]}
        for v in stuff.get("DeleteMarkers", [])
    ]
    if Delete["Objects"]:
        ret = delete_objects(
            Bucket,
            Delete,
            MFA=MFA,
            RequestPayer=RequestPayer,
            region=region,
            key=key,
            keyid=keyid,
            profile=profile,
        )
        failed = ret.get("failed", [])
        if failed:
            return {"deleted": False, "failed": ret[failed]}
    return {"deleted": True}


def list(region=None, key=None, keyid=None, profile=None):
    """
    List all buckets owned by the authenticated sender of the request.

    Returns list of buckets

    CLI Example:

    .. code-block:: yaml

        Owner: {...}
        Buckets:
          - {...}
          - {...}

    """
    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        buckets = conn.list_buckets()
        if not bool(buckets.get("Buckets")):
            log.warning("No buckets found")
        if "ResponseMetadata" in buckets:
            del buckets["ResponseMetadata"]
        return buckets
    except ClientError as e:
        return {"error": __utils__["boto3.get_error"](e)}


def list_object_versions(
    Bucket,
    Delimiter=None,
    EncodingType=None,
    Prefix=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    List objects in a given S3 bucket.

    Returns a list of objects.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_s3_bucket.list_object_versions mybucket

    """

    try:
        Versions = []
        DeleteMarkers = []
        args = {"Bucket": Bucket}
        args.update({"Delimiter": Delimiter}) if Delimiter else None
        args.update({"EncodingType": EncodingType}) if Delimiter else None
        args.update({"Prefix": Prefix}) if Prefix else None
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        IsTruncated = True
        while IsTruncated:
            ret = conn.list_object_versions(**args)
            IsTruncated = ret.get("IsTruncated", False)
            if IsTruncated in ("True", "true", True):
                args["KeyMarker"] = ret["NextKeyMarker"]
                args["VersionIdMarker"] = ret["NextVersionIdMarker"]
            Versions += ret.get("Versions", [])
            DeleteMarkers += ret.get("DeleteMarkers", [])
        return {"Versions": Versions, "DeleteMarkers": DeleteMarkers}
    except ClientError as e:
        return {"error": __utils__["boto3.get_error"](e)}


def list_objects(
    Bucket,
    Delimiter=None,
    EncodingType=None,
    Prefix=None,
    FetchOwner=False,
    StartAfter=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    List objects in a given S3 bucket.

    Returns a list of objects.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_s3_bucket.list_objects mybucket

    """

    try:
        Contents = []
        args = {"Bucket": Bucket, "FetchOwner": FetchOwner}
        args.update({"Delimiter": Delimiter}) if Delimiter else None
        args.update({"EncodingType": EncodingType}) if Delimiter else None
        args.update({"Prefix": Prefix}) if Prefix else None
        args.update({"StartAfter": StartAfter}) if StartAfter else None
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        IsTruncated = True
        while IsTruncated:
            ret = conn.list_objects_v2(**args)
            IsTruncated = ret.get("IsTruncated", False)
            if IsTruncated in ("True", "true", True):
                args["ContinuationToken"] = ret["NextContinuationToken"]
            Contents += ret.get("Contents", [])
        return {"Contents": Contents}
    except ClientError as e:
        return {"error": __utils__["boto3.get_error"](e)}


def put_acl(
    Bucket,
    ACL=None,
    AccessControlPolicy=None,
    GrantFullControl=None,
    GrantRead=None,
    GrantReadACP=None,
    GrantWrite=None,
    GrantWriteACP=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Given a valid config, update the ACL for a bucket.

    Returns {updated: true} if the ACL was updated and returns
    {updated: False} if the ACL was not updated.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_s3_bucket.put_acl my_bucket 'public' \\
                         GrantFullControl='emailaddress=example@example.com' \\
                         GrantRead='uri="http://acs.amazonaws.com/groups/global/AllUsers"' \\
                         GrantReadACP='emailaddress="exampl@example.com",id="2345678909876432"'

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        kwargs = {}
        if AccessControlPolicy is not None:
            if isinstance(AccessControlPolicy, str):
                AccessControlPolicy = salt.utils.json.loads(AccessControlPolicy)
            kwargs["AccessControlPolicy"] = AccessControlPolicy
        for arg in (
            "ACL",
            "GrantFullControl",
            "GrantRead",
            "GrantReadACP",
            "GrantWrite",
            "GrantWriteACP",
        ):
            if locals()[arg] is not None:
                kwargs[arg] = str(locals()[arg])
        conn.put_bucket_acl(Bucket=Bucket, **kwargs)
        return {"updated": True, "name": Bucket}
    except ClientError as e:
        return {"updated": False, "error": __utils__["boto3.get_error"](e)}


def put_cors(Bucket, CORSRules, region=None, key=None, keyid=None, profile=None):
    """
    Given a valid config, update the CORS rules for a bucket.

    Returns {updated: true} if CORS was updated and returns
    {updated: False} if CORS was not updated.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_s3_bucket.put_cors my_bucket '[{\\
              "AllowedHeaders":[],\\
              "AllowedMethods":["GET"],\\
              "AllowedOrigins":["*"],\\
              "ExposeHeaders":[],\\
              "MaxAgeSeconds":123,\\
        }]'

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        if CORSRules is not None and isinstance(CORSRules, str):
            CORSRules = salt.utils.json.loads(CORSRules)
        conn.put_bucket_cors(Bucket=Bucket, CORSConfiguration={"CORSRules": CORSRules})
        return {"updated": True, "name": Bucket}
    except ClientError as e:
        return {"updated": False, "error": __utils__["boto3.get_error"](e)}


def put_lifecycle_configuration(
    Bucket, Rules, region=None, key=None, keyid=None, profile=None
):
    """
    Given a valid config, update the Lifecycle rules for a bucket.

    Returns {updated: true} if Lifecycle was updated and returns
    {updated: False} if Lifecycle was not updated.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_s3_bucket.put_lifecycle_configuration my_bucket '[{\\
              "Expiration": {...},\\
              "ID": "idstring",\\
              "Prefix": "prefixstring",\\
              "Status": "enabled",\\
              "Transitions": [{...},],\\
              "NoncurrentVersionTransitions": [{...},],\\
              "NoncurrentVersionExpiration": {...},\\
        }]'

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        if Rules is not None and isinstance(Rules, str):
            Rules = salt.utils.json.loads(Rules)
        conn.put_bucket_lifecycle_configuration(
            Bucket=Bucket, LifecycleConfiguration={"Rules": Rules}
        )
        return {"updated": True, "name": Bucket}
    except ClientError as e:
        return {"updated": False, "error": __utils__["boto3.get_error"](e)}


def put_logging(
    Bucket,
    TargetBucket=None,
    TargetPrefix=None,
    TargetGrants=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Given a valid config, update the logging parameters for a bucket.

    Returns {updated: true} if parameters were updated and returns
    {updated: False} if parameters were not updated.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_s3_bucket.put_logging my_bucket log_bucket '[{...}]' prefix

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        logstate = {}
        targets = {
            "TargetBucket": TargetBucket,
            "TargetGrants": TargetGrants,
            "TargetPrefix": TargetPrefix,
        }
        for key, val in targets.items():
            if val is not None:
                logstate[key] = val
        if logstate:
            logstatus = {"LoggingEnabled": logstate}
        else:
            logstatus = {}
        if TargetGrants is not None and isinstance(TargetGrants, str):
            TargetGrants = salt.utils.json.loads(TargetGrants)
        conn.put_bucket_logging(Bucket=Bucket, BucketLoggingStatus=logstatus)
        return {"updated": True, "name": Bucket}
    except ClientError as e:
        return {"updated": False, "error": __utils__["boto3.get_error"](e)}


def put_notification_configuration(
    Bucket,
    TopicConfigurations=None,
    QueueConfigurations=None,
    LambdaFunctionConfigurations=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Given a valid config, update the notification parameters for a bucket.

    Returns {updated: true} if parameters were updated and returns
    {updated: False} if parameters were not updated.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_s3_bucket.put_notification_configuration my_bucket
                [{...}] \\
                [{...}] \\
                [{...}]

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        if TopicConfigurations is None:
            TopicConfigurations = []
        elif isinstance(TopicConfigurations, str):
            TopicConfigurations = salt.utils.json.loads(TopicConfigurations)
        if QueueConfigurations is None:
            QueueConfigurations = []
        elif isinstance(QueueConfigurations, str):
            QueueConfigurations = salt.utils.json.loads(QueueConfigurations)
        if LambdaFunctionConfigurations is None:
            LambdaFunctionConfigurations = []
        elif isinstance(LambdaFunctionConfigurations, str):
            LambdaFunctionConfigurations = salt.utils.json.loads(
                LambdaFunctionConfigurations
            )
        # TODO allow the user to use simple names & substitute ARNs for those names
        conn.put_bucket_notification_configuration(
            Bucket=Bucket,
            NotificationConfiguration={
                "TopicConfigurations": TopicConfigurations,
                "QueueConfigurations": QueueConfigurations,
                "LambdaFunctionConfigurations": LambdaFunctionConfigurations,
            },
        )
        return {"updated": True, "name": Bucket}
    except ClientError as e:
        return {"updated": False, "error": __utils__["boto3.get_error"](e)}


def put_policy(Bucket, Policy, region=None, key=None, keyid=None, profile=None):
    """
    Given a valid config, update the policy for a bucket.

    Returns {updated: true} if policy was updated and returns
    {updated: False} if policy was not updated.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_s3_bucket.put_policy my_bucket {...}

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        if Policy is None:
            Policy = "{}"
        elif not isinstance(Policy, str):
            Policy = salt.utils.json.dumps(Policy)
        conn.put_bucket_policy(Bucket=Bucket, Policy=Policy)
        return {"updated": True, "name": Bucket}
    except ClientError as e:
        return {"updated": False, "error": __utils__["boto3.get_error"](e)}


def _get_role_arn(name, region=None, key=None, keyid=None, profile=None):
    if name.startswith("arn:aws:iam:"):
        return name

    account_id = __salt__["boto_iam.get_account_id"](
        region=region, key=key, keyid=keyid, profile=profile
    )
    if profile and "region" in profile:
        region = profile["region"]
    if region is None:
        region = "us-east-1"
    return f"arn:aws:iam::{account_id}:role/{name}"


def put_replication(
    Bucket, Role, Rules, region=None, key=None, keyid=None, profile=None
):
    """
    Given a valid config, update the replication configuration for a bucket.

    Returns {updated: true} if replication configuration was updated and returns
    {updated: False} if replication configuration was not updated.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_s3_bucket.put_replication my_bucket my_role [...]

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        Role = _get_role_arn(
            name=Role, region=region, key=key, keyid=keyid, profile=profile
        )
        if Rules is None:
            Rules = []
        elif isinstance(Rules, str):
            Rules = salt.utils.json.loads(Rules)
        conn.put_bucket_replication(
            Bucket=Bucket, ReplicationConfiguration={"Role": Role, "Rules": Rules}
        )
        return {"updated": True, "name": Bucket}
    except ClientError as e:
        return {"updated": False, "error": __utils__["boto3.get_error"](e)}


def put_request_payment(Bucket, Payer, region=None, key=None, keyid=None, profile=None):
    """
    Given a valid config, update the request payment configuration for a bucket.

    Returns {updated: true} if request payment configuration was updated and returns
    {updated: False} if request payment configuration was not updated.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_s3_bucket.put_request_payment my_bucket Requester

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        conn.put_bucket_request_payment(
            Bucket=Bucket, RequestPaymentConfiguration={"Payer": Payer}
        )
        return {"updated": True, "name": Bucket}
    except ClientError as e:
        return {"updated": False, "error": __utils__["boto3.get_error"](e)}


def put_tagging(Bucket, region=None, key=None, keyid=None, profile=None, **kwargs):
    """
    Given a valid config, update the tags for a bucket.

    Returns {updated: true} if tags were updated and returns
    {updated: False} if tags were not updated.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_s3_bucket.put_tagging my_bucket my_role [...]

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        tagslist = []
        for k, v in kwargs.items():
            if str(k).startswith("__"):
                continue
            tagslist.append({"Key": str(k), "Value": str(v)})
        conn.put_bucket_tagging(Bucket=Bucket, Tagging={"TagSet": tagslist})
        return {"updated": True, "name": Bucket}
    except ClientError as e:
        return {"updated": False, "error": __utils__["boto3.get_error"](e)}


def put_versioning(
    Bucket,
    Status,
    MFADelete=None,
    MFA=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Given a valid config, update the versioning configuration for a bucket.

    Returns {updated: true} if versioning configuration was updated and returns
    {updated: False} if versioning configuration was not updated.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_s3_bucket.put_versioning my_bucket Enabled

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        VersioningConfiguration = {"Status": Status}
        if MFADelete is not None:
            VersioningConfiguration["MFADelete"] = MFADelete
        kwargs = {}
        if MFA is not None:
            kwargs["MFA"] = MFA
        conn.put_bucket_versioning(
            Bucket=Bucket, VersioningConfiguration=VersioningConfiguration, **kwargs
        )
        return {"updated": True, "name": Bucket}
    except ClientError as e:
        return {"updated": False, "error": __utils__["boto3.get_error"](e)}


def put_website(
    Bucket,
    ErrorDocument=None,
    IndexDocument=None,
    RedirectAllRequestsTo=None,
    RoutingRules=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Given a valid config, update the website configuration for a bucket.

    Returns {updated: true} if website configuration was updated and returns
    {updated: False} if website configuration was not updated.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_s3_bucket.put_website my_bucket IndexDocument='{"Suffix":"index.html"}'

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        WebsiteConfiguration = {}
        for key in (
            "ErrorDocument",
            "IndexDocument",
            "RedirectAllRequestsTo",
            "RoutingRules",
        ):
            val = locals()[key]
            if val is not None:
                if isinstance(val, str):
                    WebsiteConfiguration[key] = salt.utils.json.loads(val)
                else:
                    WebsiteConfiguration[key] = val
        conn.put_bucket_website(
            Bucket=Bucket, WebsiteConfiguration=WebsiteConfiguration
        )
        return {"updated": True, "name": Bucket}
    except ClientError as e:
        return {"updated": False, "error": __utils__["boto3.get_error"](e)}


def delete_cors(Bucket, region=None, key=None, keyid=None, profile=None):
    """
    Delete the CORS configuration for the given bucket

    Returns {deleted: true} if CORS was deleted and returns
    {deleted: False} if CORS was not deleted.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_s3_bucket.delete_cors my_bucket

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        conn.delete_bucket_cors(Bucket=Bucket)
        return {"deleted": True, "name": Bucket}
    except ClientError as e:
        return {"deleted": False, "error": __utils__["boto3.get_error"](e)}


def delete_lifecycle_configuration(
    Bucket, region=None, key=None, keyid=None, profile=None
):
    """
    Delete the lifecycle configuration for the given bucket

    Returns {deleted: true} if Lifecycle was deleted and returns
    {deleted: False} if Lifecycle was not deleted.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_s3_bucket.delete_lifecycle_configuration my_bucket

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        conn.delete_bucket_lifecycle(Bucket=Bucket)
        return {"deleted": True, "name": Bucket}
    except ClientError as e:
        return {"deleted": False, "error": __utils__["boto3.get_error"](e)}


def delete_policy(Bucket, region=None, key=None, keyid=None, profile=None):
    """
    Delete the policy from the given bucket

    Returns {deleted: true} if policy was deleted and returns
    {deleted: False} if policy was not deleted.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_s3_bucket.delete_policy my_bucket

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        conn.delete_bucket_policy(Bucket=Bucket)
        return {"deleted": True, "name": Bucket}
    except ClientError as e:
        return {"deleted": False, "error": __utils__["boto3.get_error"](e)}


def delete_replication(Bucket, region=None, key=None, keyid=None, profile=None):
    """
    Delete the replication config from the given bucket

    Returns {deleted: true} if replication configuration was deleted and returns
    {deleted: False} if replication configuration was not deleted.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_s3_bucket.delete_replication my_bucket

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        conn.delete_bucket_replication(Bucket=Bucket)
        return {"deleted": True, "name": Bucket}
    except ClientError as e:
        return {"deleted": False, "error": __utils__["boto3.get_error"](e)}


def delete_tagging(Bucket, region=None, key=None, keyid=None, profile=None):
    """
    Delete the tags from the given bucket

    Returns {deleted: true} if tags were deleted and returns
    {deleted: False} if tags were not deleted.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_s3_bucket.delete_tagging my_bucket

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        conn.delete_bucket_tagging(Bucket=Bucket)
        return {"deleted": True, "name": Bucket}
    except ClientError as e:
        return {"deleted": False, "error": __utils__["boto3.get_error"](e)}


def delete_website(Bucket, region=None, key=None, keyid=None, profile=None):
    """
    Remove the website configuration from the given bucket

    Returns {deleted: true} if website configuration was deleted and returns
    {deleted: False} if website configuration was not deleted.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_s3_bucket.delete_website my_bucket

    """

    try:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        conn.delete_bucket_website(Bucket=Bucket)
        return {"deleted": True, "name": Bucket}
    except ClientError as e:
        return {"deleted": False, "error": __utils__["boto3.get_error"](e)}
